# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import shlex
import socket
import base64
import hashlib
import binascii
from subprocess import Popen, PIPE
from threading import Event
from collections.abc import Callable, Sequence
import paramiko

from xpra.net.bytestreams import pretty_socket
from xpra.util.str_fn import csv, decode_str
from xpra.util.env import envint, osexpand, first_time
from xpra.os_util import getuid, WIN32, POSIX
from xpra.util.thread import start_thread
from xpra.scripts.config import str_to_bool
from xpra.common import SSH_AGENT_DISPATCH, SizedBuffer
from xpra.platform.paths import get_ssh_conf_dirs, get_xpra_command
from xpra.log import Logger

log = Logger("network", "ssh")

BANNER = os.environ.get("XPRA_SSH_BANNER", "Xpra SSH Server")
SERVER_WAIT = envint("XPRA_SSH_SERVER_WAIT", 20)
AUTHORIZED_KEYS = "~/.ssh/authorized_keys"
AUTHORIZED_KEYS_HASHES = os.environ.get("XPRA_AUTHORIZED_KEYS_HASHES",
                                        "md5,sha1,sha224,sha256,sha384,sha512").split(",")


def get_keyclass(keytype: str):
    if not keytype:
        return None
    # 'dsa' -> 'DSS'
    if keytype == "dsa" and hasattr(paramiko, "DSSKey"):
        return paramiko.DSSKey
    keyclass = getattr(paramiko, keytype.upper() + "Key", None)
    if keyclass:
        return keyclass
    keyclass = getattr(paramiko, keytype.upper() + "Key", None)
    if keyclass:
        return keyclass
    # Ed25519Key
    return getattr(paramiko, keytype[:1].upper() + keytype[1:] + "Key", None)


def detect_ssh_stanza(cmd: list[str]) -> Sequence[str]:
    # plain 'ssh' clients execute a long command with if+else statements,
    # try to detect it and extract the actual command the client is trying to run.
    # ie:
    # ['sh', '-c',
    # ': run-xpra _proxy;xpra initenv;\
    #  if [ -x $XDG_RUNTIME_DIR/xpra/run-xpra ]; then $XDG_RUNTIME_DIR/xpra/run-xpra _proxy;\
    #  elif [ -x ~/.xpra/run-xpra ]; then ~/.xpra/run-xpra _proxy;\
    #  elif type "xpra" > /dev/null 2>&1; then xpra _proxy;\
    #  elif [ -x /usr/local/bin/xpra ]; then /usr/local/bin/xpra _proxy;\
    #  else echo "no run-xpra command found"; exit 1; fi']
    # if .* ; then .*/run-xpra _proxy;
    log(f"parse cmd={cmd} (len={len(cmd)})")
    if len(cmd) == 1:  # ie: 'thelongcommand'
        parse_cmd = cmd[0]
    elif len(cmd) == 3 and cmd[:2] == ["sh", "-c"]:  # ie: 'sh' '-c' 'thelongcommand'
        parse_cmd = cmd[2]
    else:
        return ()
    # for older clients, try to parse the long command
    # and identify the subcommands from there
    subcommands: dict[str, list[str]] = {}
    ifparts = parse_cmd.split("if ")
    log(f"ifparts={ifparts}")
    for s in ifparts:
        if any(s.startswith(x) for x in (
                "type \"xpra\"", "which \"xpra\"", "command -v \"xpra\"", "[ -x ")
               ) and s.find("then ") > 0:
            then_str = s.split("then ", 1)[1]
            # ie: then_str="$XDG_RUNTIME_DIR/xpra/run-xpra _proxy; el"
            if then_str.find(";") > 0:
                then_str = then_str.split(";")[0]
            parts = shlex.split(then_str)
            log(f"parts({then_str})={parts}")
            if len(parts) >= 2:
                subcommand = parts[1]  # ie: "_proxy"
                if subcommand not in subcommands:
                    subcommands[subcommand] = parts
    log(f"subcommands={subcommands}")
    return subcommands.get("_proxy", ())


def find_fingerprint(filename: str, fingerprint) -> bool:
    hex_fingerprint = binascii.hexlify(fingerprint)
    log(f"looking for key fingerprint {hex_fingerprint!r} in {filename!r}")
    count = 0
    with open(filename, encoding="latin1") as f:
        for line in f:
            if line.startswith("#"):
                continue
            line = line.strip("\n\r")
            try:
                key = base64.b64decode(line.strip().split()[1].encode('ascii'))
            except Exception as e:
                log(f"ignoring line {line}: {e}")
                continue
            for hash_algo in AUTHORIZED_KEYS_HASHES:
                try:
                    hash_class = getattr(hashlib, hash_algo)  # ie: hashlib.md5
                    hash_instance = hash_class(key)  # can raise ValueError (ie: on FIPS compliant systems)
                except (AttributeError, ValueError):
                    hash_instance = None
                if not hash_instance:
                    if first_time(f"hash-{hash_algo}-missing"):
                        log.warn(f"Warning: unsupported hash {hash_algo!r}")
                        log.warn(f" in {filename!r}")
                    continue
                fp_plain = hash_instance.hexdigest()
                log(f"{hash_algo}({line})={fp_plain}")
                if fp_plain == hex_fingerprint:
                    return True
            count += 1
    log(f"no match in {count} keys from {filename!r}")
    return False


def proxy_start(channel, subcommand: str, args: list[str]) -> None:
    log(f"ssh proxy-start({channel}, {subcommand}, {args})")
    if subcommand == "_proxy_shadow_start":
        server_mode = "shadow"
    else:
        # ie: "_proxy_start_desktop" -> "start-desktop"
        server_mode = subcommand.replace("_proxy_", "").replace("_", "-")
    log.info(f"ssh channel starting proxy {server_mode} session")
    cmd = get_xpra_command() + [subcommand] + args
    try:
        # pylint: disable=consider-using-with
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, bufsize=0, close_fds=True)
        proc.poll()
    except OSError:
        log.error(f"Error starting proxy subcommand `{subcommand}`", exc_info=True)
        log.error(f" with args={args}")
        return
    # pylint: disable=import-outside-toplevel
    from xpra.util.child_reaper import getChildReaper

    def proxy_ended(*args) -> None:
        log("proxy_ended(%s)", args)

    def close() -> None:
        if proc.poll() is None:
            proc.terminate()

    getChildReaper().add_process(proc, f"proxy-start-{subcommand}", cmd, True, True, proxy_ended)

    def proc_to_channel(read: Callable[[int], SizedBuffer], send: Callable[[SizedBuffer], int]) -> None:
        while proc.poll() is None:
            # log("proc_to_channel(%s, %s) waiting for data", read, send)
            try:
                r = read(4096)
            except paramiko.buffered_pipe.PipeTimeout:
                log(f"proc_to_channel({read}, {send})", exc_info=True)
                close()
                return
            # log("proc_to_channel(%s, %s) %i bytes: %s", read, send, len(r or b""), ellipsizer(r))
            while r:
                try:
                    sent = send(r)
                    r = r[sent:]
                except OSError:
                    log(f"proc_to_channel({read}, {send})", exc_info=True)
                    close()
                    return

    # forward to/from the process and the channel:

    def stderr_reader() -> None:
        proc_to_channel(proc.stderr.read, channel.send_stderr)

    def stdout_reader() -> None:
        proc_to_channel(proc.stdout.read, channel.send)

    def stdin_reader() -> None:
        # read from channel, write to stdin
        stdin = proc.stdin
        while proc.poll() is None:
            r = channel.recv(4096)
            if not r:
                close()
                break
            stdin.write(r)
            stdin.flush()

    tname = subcommand.replace("_proxy_", "proxy-").replace("_", "-")
    start_thread(stderr_reader, f"{tname}-stderr", True)
    start_thread(stdout_reader, f"{tname}-stdout", True)
    start_thread(stdin_reader, f"{tname}-stdin", True)
    channel.proxy_process = proc


class SSHServer(paramiko.ServerInterface):
    def __init__(self, none_auth=False, pubkey_auth=True, password_auth=None, options=None, display_name=""):
        self.event = Event()
        self.none_auth = none_auth
        self.pubkey_auth = pubkey_auth
        self.password_auth = password_auth
        self.proxy_channel = None
        self.options = options or {}
        self.display_name = display_name
        self.agent = None
        self.transport = None

    def get_banner(self) -> tuple[str, str]:
        return f"{BANNER}\n\r", "EN"

    def get_allowed_auths(self, username: str) -> str:
        # return "gssapi-keyex,gssapi-with-mic,password,publickey"
        mods = []
        if self.none_auth:
            mods.append("none")
        if self.pubkey_auth:
            mods.append("publickey")
        if self.password_auth:
            mods.append("password")
        log("get_allowed_auths(%s)=%s", username, mods)
        return ",".join(mods)

    def check_channel_request(self, kind: str, chanid) -> int:
        log("check_channel_request(%s, %s)", kind, chanid)
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_channel_forward_agent_request(self, channel) -> bool:
        log(f"check_channel_forward_agent_request({channel}) SSH_AGENT_DISPATCH={SSH_AGENT_DISPATCH}")
        if SSH_AGENT_DISPATCH:
            # pylint: disable=import-outside-toplevel
            from paramiko.agent import AgentServerProxy
            self.agent = AgentServerProxy(self.transport)
            ssh_auth_sock = self.agent.get_env().get("SSH_AUTH_SOCK")
            log(f"agent SSH_AUTH_SOCK={ssh_auth_sock}")
            return bool(ssh_auth_sock)
        return False

    def check_auth_none(self, username: str) -> int:
        log("check_auth_none(%s) none_auth=%s", username, self.none_auth)
        if self.none_auth:
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def check_auth_password(self, username: str, password: str) -> int:
        log("check_auth_password(%s, %s) password_auth=%s", username, "*" * len(password), self.password_auth)
        if not self.password_auth or not self.password_auth(username, password):
            return paramiko.AUTH_FAILED
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_publickey(self, username: str, key) -> int:
        log("check_auth_publickey(%s, %r) pubkey_auth=%s", username, key, self.pubkey_auth)
        if not self.pubkey_auth:
            return paramiko.AUTH_FAILED
        if not POSIX or getuid() != 0:
            # pylint: disable=import-outside-toplevel
            import getpass
            sysusername = getpass.getuser()
            if sysusername != username:
                log.warn("Warning: ssh password authentication failed,")
                log.warn(" username does not match:")
                log.warn(f" expected {sysusername!r}, got {username!r}")
                return paramiko.AUTH_FAILED
        authorized_keys_filename = osexpand(AUTHORIZED_KEYS)
        if not os.path.exists(authorized_keys_filename) or not os.path.isfile(authorized_keys_filename):
            log("file '%s' does not exist", authorized_keys_filename)
            return paramiko.AUTH_FAILED
        fingerprint = key.get_fingerprint()
        if find_fingerprint(authorized_keys_filename, fingerprint):
            return paramiko.OPEN_SUCCEEDED
        return paramiko.AUTH_FAILED

    def check_auth_gssapi_keyex(self, username: str, gss_authenticated=paramiko.AUTH_FAILED, cc_file=None) -> int:
        log("check_auth_gssapi_keyex%s", (username, gss_authenticated, cc_file))
        return paramiko.AUTH_FAILED

    def check_auth_gssapi_with_mic(self, username: str, gss_authenticated=paramiko.AUTH_FAILED, cc_file=None) -> int:
        log("check_auth_gssapi_with_mic%s", (username, gss_authenticated, cc_file))
        return paramiko.AUTH_FAILED

    def check_channel_shell_request(self, channel) -> bool:
        log(f"check_channel_shell_request({channel})")
        return False

    def check_channel_exec_request(self, channel, command: str) -> bool:
        def fail(*messages) -> bool:
            for m in messages:
                log.warn(m)
            self.event.set()
            channel.close()
            return False

        def setup_agent(cmd) -> None:
            if SSH_AGENT_DISPATCH and self.agent:
                auth_sock = self.agent.get_env().get("SSH_AUTH_SOCK")
                log(f"paramiko agent socket={auth_sock!r}")
                if auth_sock:
                    # pylint: disable=import-outside-toplevel
                    from xpra.net.ssh.agent import setup_proxy_ssh_socket
                    setup_proxy_ssh_socket(cmd, auth_sock=auth_sock)

        def csend(exit_status=0, out=None, err=None) -> bool:
            channel.exec_response = (exit_status, out, err)
            self.event.set()
            return True

        log(f"check_channel_exec_request({channel}, {command!r})")
        cmd = shlex.split(decode_str(command))
        log(f"check_channel_exec_request: cmd={cmd}")
        if cmd[0] == "command" and len(cmd) == 1:
            return csend(out="\r\n")
        if cmd == ["command", "-v", "xpra"]:
            return csend(out="xpra\r\n")
        if cmd[0] == "ver" and len(cmd) == 1:
            # older xpra versions run this command to detect win32:
            if WIN32:
                return csend(out="Microsoft Windows")
            return csend(1, err=f"{cmd[0]}: not found\r\n")
        if cmd[0] == "echo" and len(cmd) == 2:
            # echo can be used to detect the platform,
            # so emulate a basic echo command,
            # just enough for the os detection to work:
            echo = cmd[1]
            if WIN32:
                if echo.startswith("%") and echo.endswith("%"):
                    envname = echo[1:-1]
                    if envname in ("OS", "windir"):
                        echo = os.environ.get(envname, "")
                    else:
                        echo = ""
            else:
                if echo.startswith("$"):
                    envname = echo[1:]
                    if envname in ("OSTYPE",):
                        echo = sys.platform
                    else:
                        echo = ""
            echo += "\r\n"
            log(f"exec request {cmd} returning: {echo!r}")
            return csend(out=echo)
        if WIN32 and (" ".join(cmd)).find("REG QUERY") >= 0 and str(cmd).find(
                r"HKEY_LOCAL_MACHINE\\Software\\Xpra") > 0:
            # this batch command is used to detect the xpra.exe installation path
            # (see xpra/net/ssh.py)
            # pylint: disable=import-outside-toplevel
            from xpra.platform.paths import get_app_dir
            return csend(out=f"InstallPath {get_app_dir()}\r\n")
        if cmd[0] in ("type", "which") and len(cmd) == 2:
            xpra_cmd = cmd[-1]  # ie: $XDG_RUNTIME_DIR/xpra/run-xpra or "xpra"
            # only allow '*xpra' commands:
            if any(xpra_cmd.lower().endswith(x) for x in ("xpra", "run-xpra", "xpra_cmd.exe", "xpra.exe")):
                # we don't really allow the xpra command to be executed anyway,
                # so just reply that it exists:
                return csend(out="xpra\r\n")
            return csend(1, err=f"type: {xpra_cmd!r}: not found\r\n")
        if (cmd[0].endswith("xpra") or cmd[0].endswith("Xpra_cmd.exe")) and len(cmd) >= 2:
            subcommand = cmd[1].strip("\"'").rstrip(";")
            log(f"ssh xpra subcommand: {subcommand}")
            if subcommand.startswith("_proxy_"):
                proxy_start_enabled = str_to_bool(self.options.get("proxy-start"), False)
                if not proxy_start_enabled:
                    return fail(f"Warning: received a {subcommand!r} session request",
                                " this feature is not enabled with the builtin ssh server")
                proxy_start(channel, subcommand, cmd[2:])
            elif subcommand == "_proxy":
                display_name = getattr(self, "display_name", "")
                # if specified, the display name must match this session:
                display = ""
                for arg in cmd[2:]:
                    if not arg.startswith("--"):
                        display = arg
                if display and display_name != display:
                    return fail(f"Warning: the display requested {display!r}",
                                f" does not match the current display {display_name!r}")
                log(f"ssh 'xpra {subcommand}' subcommand: display_name={display_name!r}, agent={self.agent}")
                setup_agent(cmd)
            else:
                return fail(f"Warning: unsupported xpra subcommand '{cmd[1]}'")
            # we're ready to use this socket as an xpra channel
            self._run_proxy(channel)
            return True
        proxy_cmd = detect_ssh_stanza(cmd)
        if proxy_cmd:
            setup_agent(proxy_cmd)
            self._run_proxy(channel)
            return True
        return fail("Warning: unsupported ssh command:", f" {cmd!r}")

    def _run_proxy(self, channel) -> None:
        pc = self.proxy_channel
        log(f"run_proxy({channel}) proxy-channel={pc}")
        if pc:
            self.proxy_channel = None
            pc.close()
        self.proxy_channel = channel
        self.event.set()

    def check_channel_pty_request(self, channel, term, width: int, height: int,
                                  pixelwidth: int, pixelheight: int, modes) -> bool:
        log("check_channel_pty_request%s", (channel, term, width, height, pixelwidth, pixelheight, modes))
        # refusing to open a pty:
        return False

    def enable_auth_gssapi(self) -> bool:
        log("enable_auth_gssapi()")
        return False


def make_ssh_server_connection(conn, socket_options: dict,
                               none_auth: bool = False,
                               password_auth: Callable | None = None,
                               display_name: str = ""):
    log("make_ssh_server_connection%s", (conn, socket_options, none_auth, password_auth))
    ssh_server = SSHServer(none_auth=none_auth, password_auth=password_auth, options=socket_options,
                           display_name=display_name)
    DoGSSAPIKeyExchange = str_to_bool(socket_options.get("ssh-gss-key-exchange", False), False)
    sock = conn._socket
    t = None

    def close() -> None:
        if t:
            log(f"close() closing {t}")
            try:
                t.close()
            except Exception:
                log(f"{t}.close()", exc_info=True)
        log(f"close() closing {conn}")
        try:
            conn.close()
        except Exception:
            log(f"{conn}.close()")

    try:
        t = paramiko.Transport(sock, gss_kex=DoGSSAPIKeyExchange)
        ssh_server.transport = t
        gss_host = socket_options.get("ssh-gss-host", socket.getfqdn(""))
        t.set_gss_host(gss_host)
        # load host keys:
        PREFIX = "ssh_host_"
        SUFFIX = "_key"
        host_keys = {}

        def add_host_key(fd: str, f: str) -> bool:
            ff = os.path.join(fd, f)
            keytype = f[len(PREFIX):-len(SUFFIX)]
            keyclass = get_keyclass(keytype)
            if not keyclass:
                log.warn(f"Warning: unknown host key format '{f}'")
                log(f"key type {f!r} is not supported, cannot load {ff!r}")
                return False
            log(f"loading {keytype} key from {ff!r} using {keyclass}")
            try:
                host_key = keyclass(filename=ff)
                if host_key not in host_keys:
                    host_keys[host_key] = ff
                    t.add_server_key(host_key)
                    return True
            except OSError:
                log(f"cannot add host key {ff!r}", exc_info=True)
            except paramiko.SSHException as e:
                log(f"error adding host key {ff!r}", exc_info=True)
                log.error(f"Error: cannot add {keytype} host key {ff!r}:")
                log.estr(e)
            return False

        host_key: str = socket_options.get("ssh-host-key", "")
        if host_key:
            d, f = os.path.split(host_key)
            if f.startswith(PREFIX) and f.endswith(SUFFIX):
                add_host_key(d, f)
            if not host_keys:
                log.error(f"Error: failed to load host key {host_key!r}")
                close()
                return None
        else:
            ssh_key_dirs = get_ssh_conf_dirs()
            log("trying to load ssh host keys from: " + csv(ssh_key_dirs))
            for d in ssh_key_dirs:
                fd = osexpand(d)
                log(f"osexpand({d})={fd}")
                if not os.path.exists(fd) or not os.path.isdir(fd):
                    log(f"ssh host key directory {fd!r} is invalid")
                    continue
                for f in os.listdir(fd):
                    if f.startswith(PREFIX) and f.endswith(SUFFIX):
                        add_host_key(fd, f)
            if not host_keys:
                log.error("Error: cannot start SSH server,")
                log.error(" no readable SSH host keys found in:")
                log.error(" " + csv(ssh_key_dirs))
                close()
                return None
        log("loaded host keys: %s", tuple(host_keys.values()))
        t.start_server(server=ssh_server)
    except EOFError:
        log("SSH connection closed", exc_info=True)
        close()
        return None
    except paramiko.SSHException as e:
        log("failed to start ssh server", exc_info=True)
        log.error("Error handling SSH connection:")
        log.estr(e)
        close()
        return None
    while t.is_active():
        try:
            chan = t.accept(SERVER_WAIT)
            if chan is None:
                log.warn("Warning: SSH channel setup failed")
                log.warn(" closing connection %s", conn)
                # prevent errors trying to access this connection, now likely dead:
                conn.set_active(False)
                close()
                return None
        except paramiko.SSHException as e:
            log("failed to open ssh channel", exc_info=True)
            log.error("Error opening channel:")
            log.estr(e)
            close()
            return None
        log(f"client authenticated, accepted channel={chan}")
        timedout = not ssh_server.event.wait(SERVER_WAIT)
        if timedout:
            log.warn("Warning: timeout waiting for xpra SSH subcommand,")
            log.warn(" closing connection from " + pretty_socket(conn.target))
            close()
            return None
        proxy_channel = ssh_server.proxy_channel
        if proxy_channel:
            if getattr(proxy_channel, "proxy_process", None):
                log("proxy channel is handled using a subprocess")
                return None
            from xpra.net.ssh.paramiko_client import SSHSocketConnection
            return SSHSocketConnection(proxy_channel, sock,
                                       conn.local, conn.endpoint, conn.target,
                                       socket_options=socket_options)
        exec_response = getattr(chan, "exec_response", None)
        log(f"proxy channel={proxy_channel}, timedout={timedout}, exec_response={exec_response}")
        if exec_response:
            exit_status, out, err = exec_response
            if out:
                chan.sendall(out)
            if err:
                chan.sendall_stderr(err)
            chan.send_exit_status(exit_status)
            chan.close()
            # the client may now make another request on a new channel:
            ssh_server.event.clear()
    return None



# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=import-outside-toplevel

import os.path
import sys
import site
import tempfile


def do_get_desktop_background_paths():
    return [
        "/usr/share/backgrounds/images/default.png",
        "/usr/share/backgrounds/images/*default*.png",
        "/usr/share/backgrounds/*default*png",
        "/usr/share/backgrounds/gnome/adwaita*.jpg",  # Debian Stretch
        "/usr/share/backgrounds/images/*jpg",  # CentOS 7
    ]


def do_get_install_prefix():
    # special case for "user" installations, ie:
    # $HOME/.local/lib/python3.8/site-packages/xpra/platform/paths.py
    try:
        base = site.getuserbase()
    except Exception:
        base = site.USER_BASE or ""
    if __file__.startswith(base):
        return base
    if sys.argv:
        p = sys.argv[0].find("/bin/xpra")
        if p > 0:
            return sys.argv[0][:p]
    return sys.prefix


def do_get_resources_dir():
    # is there a better/cleaner way?
    from xpra.common import DEFAULT_XDG_DATA_DIRS
    from xpra.platform.paths import get_install_prefix
    options = [
        get_install_prefix(),
        sys.exec_prefix,
    ] + os.environ.get("XDG_DATA_DIRS", DEFAULT_XDG_DATA_DIRS).split(":")
    for x in options:
        p = os.path.join(x, "share", "xpra")
        if os.path.exists(p) and os.path.isdir(p):
            return p
    try:
        # test for a local installation path (run from source tree):
        local_share_path = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]), "..", "share", "xpra"))
        if os.path.exists(local_share_path) and os.path.isdir(local_share_path):
            return local_share_path
    except Exception:
        pass
    return os.getcwd()


def do_get_app_dir():
    from xpra.platform.paths import get_resources_dir
    return get_resources_dir()


def do_get_image_dir():
    from xpra.platform.paths import get_app_dir
    return os.path.join(get_app_dir(), "images")


def do_get_icon_dir():
    from xpra.platform.paths import get_app_dir
    return os.path.join(get_app_dir(), "icons")


def do_get_mmap_dir():
    return _get_xpra_runtime_dir() or tempfile.gettempdir()


def do_get_xpra_tmp_dir():
    xrd = os.environ.get("XPRA_SESSION_DIR", _get_xpra_runtime_dir())
    return os.path.join(xrd, "tmp")


def do_get_script_bin_dirs():
    # versions before 0.17 only had "~/.xpra/run-xpra"
    script_bin_dirs = []
    runtime_dir = _get_xpra_runtime_dir()
    if runtime_dir:
        script_bin_dirs.append(runtime_dir)
    return script_bin_dirs


def do_get_system_conf_dirs():
    dirs = ["/etc/xpra", "/usr/local/etc/xpra"]
    for d in os.environ.get("XDG_CONFIG_DIRS", "/etc/xdg").split(":"):
        dirs.append(os.path.join(d, "xpra"))
    # hope the prefix is something like "/usr/local" or "$HOME/.local":
    from xpra.platform.paths import get_install_prefix
    prefix = get_install_prefix()
    if prefix not in ("/usr", "/usr/local"):
        if prefix.endswith(".local"):
            idir = os.path.join(prefix, "xpra")  # ie: ~/.local/xpra
        else:
            idir = os.path.join(prefix, "/etc/xpra/")  # ie: /someinstallpath/etc/xpra
        if idir not in dirs:
            dirs.append(idir)
    return dirs


def do_get_system_menu_dirs() -> list[str]:
    from xpra.common import DEFAULT_XDG_DATA_DIRS
    data_dirs = os.environ.get("XDG_DATA_DIRS", DEFAULT_XDG_DATA_DIRS).split(":")
    menu_dirs = []
    for data_dir in data_dirs:
        menu_dir = os.path.join(os.path.expanduser(data_dir), "applications")
        if os.path.exists(menu_dir) and os.path.isdir(menu_dir):
            menu_dirs.append(menu_dir)
    return menu_dirs


def do_get_state_dir():
    if os.getuid() > 0:
        xdg_state = os.path.expanduser(os.environ.get("XDG_STATE_HOME", "~/.local/state"))
        state_dir = os.path.expanduser(xdg_state)
        if state_dir and os.path.exists(state_dir):
            xpra_state = os.path.join(state_dir, "xpra")
            if not os.path.exists(xpra_state):
                os.mkdir(xpra_state)
            return os.path.join(xdg_state, "xpra")
    return "~/.xpra/state"


def do_get_user_conf_dirs(uid):
    # per-user configuration location:
    # (but never use /root/.xpra or /root/.config/xpra)
    if uid is None:
        uid = os.getuid()
    dirs = []
    if uid > 0:
        dirs += [os.path.join(os.environ.get("XDG_CONFIG_HOME", "~/.config"), "xpra")]
        dirs.append("~/.xpra")
    return dirs


def get_runtime_dir():
    runtime_dir = os.environ.get("XDG_RUNTIME_DIR")
    if runtime_dir:
        return runtime_dir
    if sys.platform.startswith("linux"):
        for d in ("/run/user", "/var/run/user"):
            if os.path.exists(d) and os.path.isdir(d):
                runtime_dir = d + "/$UID"
                break
        if not runtime_dir:
            return "$XDG_RUNTIME_DIR"
    return runtime_dir


def _get_xpra_runtime_dir():
    runtime_dir = get_runtime_dir()
    if not runtime_dir:
        return None
    return os.path.join(runtime_dir, "xpra")


def do_get_socket_dirs():
    SOCKET_DIRS = []
    runtime_dir = _get_xpra_runtime_dir()
    if runtime_dir:
        # private, per user: XDG_RUNTIME_DIR/xpra
        # (ie: "/run/user/1000/xpra")
        SOCKET_DIRS.append(runtime_dir)
    # for shared sockets (the 'xpra' group should own this directory):
    if os.path.exists("/run"):
        SOCKET_DIRS.append("/run/xpra")
    elif os.path.exists("/var/run"):
        SOCKET_DIRS.append("/var/run/xpra")
    # Debian and Ubuntu often don't create a reliable XDG_RUNTIME_DIR
    # other distros may not create one when using "su"
    SOCKET_DIRS.append("~/.xpra")
    return SOCKET_DIRS


def do_get_client_socket_dirs():
    DIRS = []
    runtime_dir = _get_xpra_runtime_dir()
    if runtime_dir:
        DIRS.append(os.path.join(runtime_dir, "clients"))
    return DIRS


def do_get_default_log_dirs():
    log_dirs = []
    v = _get_xpra_runtime_dir()
    if v:
        log_dirs.append(v)
    log_dirs.append(tempfile.gettempdir())
    return log_dirs


def do_get_audio_command():
    from xpra.platform.paths import get_xpra_command
    return get_xpra_command()


def do_get_xpra_command():
    # try to use the same "xpra" executable that launched this server,
    # whilst also preserving the python interpreter version:
    if sys.argv and sys.argv[0].lower().endswith("/xpra"):
        if sys.executable and sys.executable.find("python") >= 0:
            return [sys.executable, sys.argv[0]]
        return ["python%i.%i" % (sys.version_info.major, sys.version_info.minor), sys.argv[0]]
    return ["xpra"]



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# https://mymusing.co/full-range-vs-narrow-range-color/
# narrow range:
# Y (aka luma) component ranges from 16 to 235
#  hence: 255/(235-16) = 1.1643835616438356
# Cb and Cr (aka U and V) range from 16 to 240
#  hence: 255/(240-16) = 1.1383928571428572

# same for the offset:
#  Y: 16/255 = 0.062745098
#  U and V: 0.5 + 0.062745098 = 0.562745098

# https://gist.github.com/yohhoy/dafa5a47dade85d8b40625261af3776a
# Y  = a * R + b * G + c * B
# Cb = (B - Y) / d
# Cr = (R - Y) / e
# BT.601	BT.709	BT.2020
# a	0.299	0.2126	0.2627
# b	0.587	0.7152	0.6780
# c	0.114	0.0722	0.0593
# d	1.772	1.8556	1.8814
# e	1.402	1.5748	1.4746
# R = Y + e * Cr
# G = Y - (a * e / b) * Cr - (c * d / b) * Cb
# B = Y + d * Cb

from xpra.codecs.constants import get_subsampling_divs

GLSL_VERSION = "330 core"

CS_MULTIPLIERS: dict[str, tuple[float, float, float, float, float]] = {
    "bt601": (0.299, 0.587, 0.114, 1.772, 1.402),
    "bt709": (0.2126, 0.7152, 0.0722, 1.8556, 1.5748),
    "bt2020": (0.2627, 0.6780, 0.0593, 1.8814, 1.4746),
}


def gen_YUV_to_RGB(fmt="YUV420P", cs="bt601", full_range=True) -> str:
    if cs not in CS_MULTIPLIERS:
        raise ValueError(f"unsupported colorspace {cs}")

    a, b, c, d, e = CS_MULTIPLIERS[cs]
    f = - c * d / b
    g = - a * e / b
    ymult = "" if full_range else " * 1.1643835616438356"
    uvmult = "" if full_range else " * 1.1383928571428572"
    yoffset = "" if full_range else " - 0.062745098"
    defines = []

    def add_div(name: str, xdiv=1, ydiv=1):
        if xdiv == ydiv:
            # just divide by the same integer:
            # ie: "#define Ydiv 1"
            value = xdiv
        else:
            # store each component in a vector:
            # ie: "#define Udiv vec2(2, 1)"
            value = f"vec2({xdiv}, {ydiv})"
        defines.append(f"{name}div {value}")

    divs = get_subsampling_divs(fmt)
    for i, div in enumerate(divs):
        add_div("YUVA"[i], *div)

    has_alpha = fmt.find("A") >= 0
    if has_alpha:
        alphasampler = "uniform sampler2DRect A;"
        alpha = "texture(A, pos/Adiv).r"
    else:
        alphasampler = ""
        alpha = "1.0"

    defines_str = "\n".join(f"#define {define}" for define in defines)
    return f"""
#version {GLSL_VERSION}
layout(origin_upper_left) in vec4 gl_FragCoord;
{defines_str}
uniform vec2 viewport_pos;
uniform vec2 scaling;
uniform sampler2DRect Y;
uniform sampler2DRect U;
uniform sampler2DRect V;
{alphasampler}
layout(location = 0) out vec4 frag_color;

void main()
{{
    vec2 pos = (gl_FragCoord.xy-viewport_pos.xy)/scaling;
    highp float y = (texture(Y, pos/Ydiv).r{yoffset}){ymult};
    highp float u = (texture(U, pos/Udiv).r - 0.5){uvmult};
    highp float v = (texture(V, pos/Vdiv).r - 0.5){uvmult};
    highp float a = {alpha};

    highp float r = y +           {e} * v;
    highp float g = y + {f} * u + {g} * v;
    highp float b = y + {d} * u;

    frag_color = vec4(r, g, b, a);
}}
"""


def gen_NV12_to_RGB(cs="bt601", full_range=True) -> str:
    if cs not in CS_MULTIPLIERS:
        raise ValueError(f"unsupported colorspace {cs}")
    a, b, c, d, e = CS_MULTIPLIERS[cs]
    f = - c * d / b
    g = - a * e / b
    ymult = "" if full_range else " * 1.1643835616438356"
    umult = vmult = "" if full_range else " * 1.1383928571428572"
    return f"""
#version {GLSL_VERSION}
layout(origin_upper_left) in vec4 gl_FragCoord;
uniform vec2 viewport_pos;
uniform vec2 scaling;
uniform sampler2DRect Y;
uniform sampler2DRect UV;
layout(location = 0) out vec4 frag_color;

void main()
{{
    vec2 pos = (gl_FragCoord.xy-viewport_pos.xy)/scaling;
    highp float y = texture(Y, pos).r {ymult};
    highp float u = (texture(UV, pos).r - 0.5) {umult};
    highp float v = (texture(UV, pos).g - 0.5) {vmult};

    highp float r = y +           {e} * v;
    highp float g = y + {f} * u + {g} * v;
    highp float b = y + {d} * u;

    frag_color = vec4(r, g, b, 1.0);
}}
"""


VERTEX_SHADER = f"""
#version {GLSL_VERSION}
layout(location=0) in vec4 position;

void main()
{{
    gl_Position = vec4(position.x, position.y, 1, 1);
}}
"""

OVERLAY_SHADER = f"""
#version {GLSL_VERSION}
layout(origin_upper_left) in vec4 gl_FragCoord;
uniform vec2 viewport_pos;
uniform sampler2DRect rgba;
layout(location = 0) out vec4 frag_color;

void main()
{{
    vec2 pos = gl_FragCoord.xy-viewport_pos.xy;
    frag_color = texture(rgba, mod(pos, textureSize(rgba)));
    if ( frag_color.a < 0.3 ) {{
        discard;
    }}
}}
"""

BLEND_SHADER = f"""
#version {GLSL_VERSION}
layout(origin_upper_left) in vec4 gl_FragCoord;
uniform vec2 viewport_pos;
uniform sampler2DRect rgba;
uniform sampler2DRect fbo;
layout(location = 0) out vec4 frag_color;

void main()
{{
    vec2 tex_pos = gl_FragCoord.xy - viewport_pos.xy;
    vec4 tex_color = texture(rgba, mod(tex_pos, textureSize(rgba)));
    vec2 fbo_pos = vec2(gl_FragCoord.x, textureSize(fbo).y - gl_FragCoord.y);
    vec4 fbo_color = texture(fbo, fbo_pos);
    frag_color = mix(tex_color, fbo_color, 0.5);
}}
"""

FIXED_COLOR_SHADER = f"""
#version {GLSL_VERSION}
uniform vec4 color;
layout(location = 0) out vec4 frag_color;

void main()
{{
    frag_color = color;
}}
"""

SOURCE: dict[str, str] = {
    "blend": BLEND_SHADER,
    "vertex": VERTEX_SHADER,
    "overlay": OVERLAY_SHADER,
    "fixed-color": FIXED_COLOR_SHADER,
}

for full in (False, True):
    suffix = "_FULL" if full else ""
    SOURCE[f"NV12_to_RGB{suffix}"] = gen_NV12_to_RGB(full_range=full)

    for fmt in (
        "YUV420P", "YUV422P", "YUV444P",
        "YUVA420P", "YUVA422P", "YUVA444P",
    ):
        SOURCE[f"{fmt}_to_RGB{suffix}"] = gen_YUV_to_RGB(fmt, full_range=full)


def main():
    for shader, source in SOURCE.items():
        print("#"*80)
        print(f"#{shader}:")
        print()
        print(source)
        print()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import TypeAlias
from collections.abc import Sequence


# These chipsets will use OpenGL,
# there will not be any warnings, even if the vendor is greylisted:
GL_MATCH_LIST: TypeAlias = dict[str, Sequence[str]]

WHITELIST: GL_MATCH_LIST = {
}

# Chipsets from these vendors will be disabled without triggering any warnings:
GREYLIST: GL_MATCH_LIST = {
    "renderer":
        (
            "SVGA3D",
            "Software Rasterizer",
            "llvmpipe",
            # "NV134",
        ),
}

# These chipsets will be disabled by default:
BLOCKLIST: GL_MATCH_LIST = {
    "renderer":
        (),
    "vendor":
        (
            # "VMware, Inc.",
            # "Humper",
            # to disable nvidia, uncomment this:
            # "NVIDIA Corporation",
        ),
    "platform":
        (
            # "darwin",
        ),
}


# for testing:
# GREYLIST["vendor"].append("NVIDIA Corporation")
# WHITELIST["renderer"] = ["GeForce GTX 760/PCIe/SSE2"]
# frequent crashes on OSX with GT 650M: (see ticket #808)
# if OSX:
#    GREYLIST.setdefault("vendor", []).append("NVIDIA Corporation")


class OpenGLFatalError(ImportError):
    pass



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from typing import Any

from xpra.scripts.config import str_to_bool
from xpra.server.mixins.stub_server_mixin import StubServerMixin


class MMAP_Server(StubServerMixin):
    """
    Mixin for servers that can handle mmap transfers
    """
    PREFIX = "mmap"

    def __init__(self):
        self.mmap_supported = False
        self.mmap_filename = ""
        self.mmap_min_size = 64 * 1024 * 1024

    def init(self, opts) -> None:
        if opts.mmap and os.path.isabs(opts.mmap):
            self.mmap_supported = True
            self.mmap_filename = opts.mmap
        else:
            self.mmap_supported = str_to_bool(opts.mmap)

    def get_info(self, _proto=None) -> dict[str, Any]:
        return {
            "mmap": {
                "supported": self.mmap_supported,
                "filename": self.mmap_filename or "",
            },
        }



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from collections.abc import Callable

import objc
from Quartz import (
    CGWindowListCopyWindowInfo, kCGWindowListOptionOnScreenOnly,
    kCGNullWindowID, kCGWindowListOptionAll,
)
from AppKit import (
    NSObject, NSWorkspace, NSApplication,
    NSWorkspaceActiveSpaceDidChangeNotification,
    NSWorkspaceWillSleepNotification,
    NSWorkspaceDidWakeNotification,
)

from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("osx", "events")
workspacelog = Logger("osx", "events", "workspace")

SLEEP_HANDLER = envbool("XPRA_OSX_SLEEP_HANDLER", True)


class AppDelegate(NSObject):

    # noinspection PyTypeHints
    def init(self) -> None:
        objc_self = objc.super(AppDelegate, self).init()
        if self is None:
            return None
        objc_self.callbacks: dict[str, list[Callable]] = {}
        objc_self.workspace = None
        objc_self.notificationCenter = None
        return objc_self

    @objc.python_method
    def register_sleep_handlers(self):
        log("register_sleep_handlers()")
        self.workspace: NSWorkspace = NSWorkspace.sharedWorkspace()
        self.notificationCenter = self.workspace.notificationCenter()

        def add_observer(fn: Callable, val) -> None:
            self.notificationCenter.addObserver_selector_name_object_(self, fn, val, None)

        # NSWorkspaceWillPowerOffNotification
        add_observer(self.receiveSleepNotification_, NSWorkspaceWillSleepNotification)
        add_observer(self.receiveWakeNotification_, NSWorkspaceDidWakeNotification)
        add_observer(self.receiveWorkspaceChangeNotification_, NSWorkspaceActiveSpaceDidChangeNotification)

    @objc.typedSelector(b'B@:#B')
    def applicationShouldHandleReopen_hasVisibleWindows_(self, ns_app, flag):
        log("applicationShouldHandleReopen_hasVisibleWindows%s", (ns_app, flag))
        self.delegate_cb("deiconify")
        return True

    @objc.typedSelector(b'v@:I')
    def receiveWorkspaceChangeNotification_(self, aNotification):
        workspacelog("receiveWorkspaceChangeNotification_(%s)", aNotification)
        if not CGWindowListCopyWindowInfo:
            return
        with workspacelog.trap_error("Error querying workspace info"):
            ourpid = os.getpid()
            # list all windows on screen:
            option = kCGWindowListOptionAll | kCGWindowListOptionOnScreenOnly
            windowList = CGWindowListCopyWindowInfo(option, kCGNullWindowID)
            our_windows = {}
            for window in windowList:
                pid = window['kCGWindowOwnerPID']
                if pid == ourpid:
                    num = window['kCGWindowNumber']
                    name = window['kCGWindowName']
                    our_windows[num] = name
            workspacelog("workspace change - our windows on screen: %s", our_windows)
            if our_windows:
                self.call_handlers("resume")
            else:
                self.call_handlers("suspend")

    @objc.typedSelector(b'v@:I')
    def receiveSleepNotification_(self, notification):
        log("receiveSleepNotification_(%s)", notification)
        self.call_handlers("suspend")

    @objc.typedSelector(b'v@:I')
    def receiveWakeNotification_(self, notification):
        log("receiveWakeNotification_(%s)", notification)
        self.call_handlers("resume")

    @objc.python_method
    def call_handlers(self, name: str) -> None:
        callbacks = self.callbacks.get(name, [])
        log("delegate_cb(%s) callbacks=%s", name, callbacks)
        for callback in callbacks:
            with log.trap_error("Error in %s callback %s", name, callback):
                callback()

    def add_handler(self, event: str, handler: Callable) -> None:
        self.callbacks.setdefault(event, []).append(handler)

    def remove_handler(self, event: str, handler: Callable) -> None:
        callbacks = self.callbacks.get(event, [])
        if handler in callbacks:
            callbacks.remove(handler)


delegate = None
shared_app = None


def get_app_delegate() -> AppDelegate:
    global delegate, shared_app
    if not delegate:
        shared_app = NSApplication.sharedApplication()
        delegate = AppDelegate.alloc()
        delegate.init()
        delegate.retain()
        if SLEEP_HANDLER:
            delegate.register_sleep_handlers()
        shared_app.setDelegate_(delegate)
        log(f"setup_event_listener() the application delegate {delegate} has been registered with {shared_app}")
    return delegate


def add_handler(event: str, handler: Callable) -> None:
    get_app_delegate().add_handler(event, handler)


def remove_handler(event: str, handler: Callable) -> None:
    get_app_delegate().remove_handler(event, handler)



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.util.env import envbool
from xpra.gtk.error import xsync, xlog
from xpra.x11.gtk.prop import prop_get
from xpra.x11.bindings.window import X11WindowBindings
from xpra.log import Logger

log = Logger("x11", "window")

WM_S0 = "WM_S0"
_NEW_WM_CM_S0 = "_NEW_WM_CM_S0"

FORCE_REPLACE_WM = envbool("XPRA_FORCE_REPLACE_WM", False)


def get_ewmh_xid() -> int:
    with xlog:
        X11Window = X11WindowBindings()
        root_xid = X11Window.get_root_xid()
        ewmh_xid = prop_get(root_xid, "_NET_SUPPORTING_WM_CHECK", "window", ignore_errors=True)
        if ewmh_xid:
            try:
                with xsync:
                    if X11Window.getGeometry(ewmh_xid):
                        return ewmh_xid
            except Exception as e:
                log(f"getGeometry({ewmh_xid:x}) {e}")
    return 0


def get_wm_info() -> dict[str, Any]:
    from xpra.x11.bindings.display_source import get_display_name
    with xsync:
        X11Window = X11WindowBindings()
        root_xid = X11Window.get_root_xid()
        info = {
            "root": root_xid,
            "display": get_display_name(),
        }
        s0 = X11Window.XGetSelectionOwner(WM_S0)
        if s0:
            info["WM_S0"] = s0
        s0 = X11Window.XGetSelectionOwner(_NEW_WM_CM_S0)
        if s0:
            info["_NEW_WM_CM_S0"] = s0
    ewmh_xid = get_ewmh_xid()
    if ewmh_xid:
        info["_NET_SUPPORTING_WM_CHECK"] = ewmh_xid
        with xlog:
            wm_name = prop_get(ewmh_xid, "_NET_WM_NAME", "utf8", ignore_errors=True)
            if not wm_name:
                wm_name = prop_get(root_xid, "_NET_WM_NAME", "utf8", ignore_errors=True)
            if wm_name:
                info["wmname"] = wm_name
    for name, prop_name, prop_type in (
            ("xpra-server-pid", "XPRA_SERVER_PID", "u32"),
            ("xpra-vfb-pid", "XPRA_XVFB_PID", "u32"),
            ("xpra-server-version", "XPRA_SERVER", "latin1"),
            ("xpra-server-mode", "XPRA_SERVER_MODE", "latin1"),
            ("dbus-address", "DBUS_SESSION_BUS_ADDRESS", "latin1"),
            ("dbus-pid", "DBUS_SESSION_BUS_PID", "u32"),
            ("dbus-window", "DBUS_SESSION_BUS_WINDOW_ID", "u32"),
    ):
        v = prop_get(root_xid, prop_name, prop_type, ignore_errors=True, raise_xerrors=False)
        if v is not None:
            info[name] = v
    log("get_wm_info()=%s", info)
    return info


def wm_check(upgrading=False) -> bool:
    info = get_wm_info()
    display_name = info.get("display", "")
    name = info.get("wmname")
    wm_so = info.get("WM_S0")
    cwm_so = info.get("_NEW_WM_CM_S0")
    ewmh_xid = info.get("_NET_SUPPORTING_WM_CHECK", 0)
    xpra_name = name and name.lower().startswith("xpra")
    if not upgrading and not (ewmh_xid or wm_so or cwm_so):
        log("no window manager on %s", display_name)
        return True
    if upgrading and xpra_name:
        log.info("found previous Xpra instance")
        return True
    if not name:
        log.warn("Warning: no window manager found")
        log.warn(" on display %s using EWMH window %#x", display_name, ewmh_xid)
    else:
        log.warn("Warning: found an existing window manager")
        log.warn(" on display %s using EWMH window %#x: '%s'", display_name, ewmh_xid, name)
    if not wm_so and not cwm_so:
        if xpra_name:
            log.info(" found remnants of a previous Xpra instance")
            return True
        if FORCE_REPLACE_WM:
            log.warn(" XPRA_FORCE_REPLACE_WM is set, replacing it forcibly")
            return True
        log.warn(" it does not own the selection '%s' or '%s'", WM_S0, _NEW_WM_CM_S0)
        log.warn(" so we cannot take over and make it exit")
        log.warn(" please stop %s so you can run xpra on this display",
                 name or "the existing window manager")
        log.warn(" if you are certain that the window manager is already gone,")
        log.warn(" you may set XPRA_FORCE_REPLACE_WM=1 to force xpra to continue")
        log.warn(" at your own risk")
        return False
    if upgrading and not FORCE_REPLACE_WM and name:
        log.error("Error: %r is managing this display", name)
        log.error(" you may set XPRA_FORCE_REPLACE_WM=1 to force xpra to continue")
        log.error(" at your own risk")
        return False
    return True


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.x11.gtk.display_source import init_gdk_display_source
    init_gdk_display_source()
    wm_check()


if __name__ == "__main__":  # pragma: no cover
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Ubuntu's "tray" is not very useful:
# - we can't know its position
# - we can't get pointer motion events
# - we can't set the icon using a path (not easily anyway)
# - we can only show a menu and nothing else
# - that menu looks bloody awful
# etc

import os
import sys
import tempfile
from time import monotonic

from xpra.util.env import envbool, osexpand
from xpra.os_util import gi_import
from xpra.client.gui.tray_base import TrayBase
from xpra.platform.paths import get_icon_dir, get_icon_filename, get_xpra_tmp_dir
from xpra.log import Logger

log = Logger("tray", "posix")

try:
    AppIndicator3 = gi_import("AyatanaAppIndicator3", "0.1")
    log("loaded `AyatanaAppIndicator3`")
except (ImportError, ValueError):
    log("failed to load `AyatanaAppIndicator3`", exc_info=True)
    try:
        AppIndicator3 = gi_import("AppIndicator3", "0.1")  # @UndefinedVariable
        log("loaded `AppIndicator3`")
    except ValueError as ve:
        log("failed to load `AppIndicator3`", exc_info=True)
        raise ImportError(f"failed to load `AppIndicator3`: {ve}") from None

DELETE_TEMP_FILE = envbool("XPRA_APPINDICATOR_DELETE_TEMP_FILE", True)

PASSIVE = AppIndicator3.IndicatorStatus.PASSIVE
ACTIVE = AppIndicator3.IndicatorStatus.ACTIVE
APPLICATION_STATUS = AppIndicator3.IndicatorCategory.APPLICATION_STATUS


def Indicator(tooltip: str, filename: str, status: AppIndicator3.IndicatorStatus):
    return AppIndicator3.Indicator.new(id=tooltip, icon_name=filename, category=status)


class AppindicatorTray(TrayBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        filename = get_icon_filename(self.default_icon_filename or "xpra.png") or "xpra.png"
        self._has_icon = False
        self.tmp_filename = ""
        self.tray_widget = Indicator(self.tooltip, filename, APPLICATION_STATUS)
        log(f"AppindicatorTray widget={self.tray_widget} for {self.tooltip=}, {filename=}, {self.menu=}")
        if hasattr(self.tray_widget, "set_icon_theme_path"):
            self.tray_widget.set_icon_theme_path(get_icon_dir())
        try:
            self.tray_widget.set_attention_icon_full("xpra.png", self.tooltip)
        except AttributeError:
            self.tray_widget.set_attention_icon("xpra.png")
        if filename:
            self.set_icon_from_file(filename)
        if not self._has_icon:
            self.tray_widget.set_label("Xpra", "")
        if self.menu:
            self.tray_widget.set_menu(self.menu)
        self.show()

    def get_geometry(self):
        # no way to tell :(
        return None

    def hide(self) -> None:
        log("Indicator.set_status(PASSIVE)")
        self.tray_widget.set_status(PASSIVE)

    def show(self) -> None:
        log("Indicator.set_status(ACTIVE)")
        self.tray_widget.set_status(ACTIVE)

    def set_blinking(self, on: bool) -> None:
        # "I'm Afraid I Can't Do That"
        pass

    def set_tooltip(self, tooltip="") -> None:
        # we only use this if we haven't got an icon
        # as with appindicator this creates a large text label
        # next to where the icon is/should be
        if not self._has_icon:
            self.tray_widget.set_label(tooltip or "Xpra", "")

    def set_icon_from_data(self, pixels, has_alpha: bool, w: int, h: int, rowstride: int, _options=None) -> None:
        # use a temporary file (yuk)
        self.clean_last_tmp_icon()
        # pylint: disable=import-outside-toplevel
        from xpra.gtk.pixbuf import pixbuf_save_to_memory
        from xpra.gtk.pixbuf import get_pixbuf_from_data
        tray_icon = get_pixbuf_from_data(pixels, has_alpha, w, h, rowstride)
        png_data = pixbuf_save_to_memory(tray_icon)
        tmp_dir = osexpand(get_xpra_tmp_dir())
        if not os.path.exists(tmp_dir):
            os.mkdir(tmp_dir, 0o755)
        fd = None
        try:
            fd, self.tmp_filename = tempfile.mkstemp(prefix="tray", suffix=".png", dir=tmp_dir)
            log("set_icon_from_data%s using temporary file %s",
                (f"{len(pixels)} pixels", has_alpha, w, h, rowstride), self.tmp_filename)
            os.write(fd, png_data)
        except OSError as e:
            log("error saving temporary file", exc_info=True)
            log.error("Error saving icon data to temporary file")
            log.estr(e)
            return
        finally:
            if fd:
                os.fchmod(fd, 0o644)
                os.close(fd)
        self.do_set_icon_from_file(self.tmp_filename)

    def do_set_icon_from_file(self, filename: str) -> None:
        if filename and os.path.exists(filename):
            self._has_icon = True
            self.icon_timestamp = monotonic()
            self.tray_widget.set_icon_full(filename, self.tooltip or "Xpra")

    def clean_last_tmp_icon(self) -> None:
        if self.tmp_filename and DELETE_TEMP_FILE:
            try:
                os.unlink(self.tmp_filename)
            except OSError:
                log("failed to remove tmp icon", exc_info=True)
            self.tmp_filename = ""

    def cleanup(self) -> None:
        self.clean_last_tmp_icon()
        super().cleanup()


def main():  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("AppIndicator-Test", "AppIndicator Test"):
        if "-v" in sys.argv:
            from xpra.log import enable_debug_for
            enable_debug_for("tray")

        from xpra.gtk.signals import register_os_signals
        Gtk = gi_import("Gtk")
        menu = Gtk.Menu()
        item = Gtk.MenuItem(label="Top Menu Item 1")
        submenu = Gtk.Menu()
        item.set_submenu(submenu)
        sub = Gtk.MenuItem(label="Sub Menu Item 1")
        subsubmenu = Gtk.Menu()
        sub.set_submenu(subsubmenu)
        for n in range(1, 1000):
            subsubmenu.append(Gtk.MenuItem(label="Sub Sub Menu Item " + str(n)))
        submenu.append(sub)
        sub = Gtk.MenuItem(label="Sub Menu Item 2")
        submenu.append(sub)
        menu.append(item)
        item = Gtk.MenuItem(label="Top Menu Item 2")
        menu.append(item)
        menu.show_all()
        a = AppindicatorTray(None, None, menu, "test", icon_filename="xpra.png", exit_cb=Gtk.main_quit)
        a.show()
        register_os_signals(Gtk.main_quit)
        Gtk.main()


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from pyglet.window import key


keynames = {}
for attr in dir(key):
    if attr.startswith("_") or not attr.isupper():
        continue
    value = getattr(key, attr)
    keynames[value] = attr



#!/usr/bin/env python3

import gi

gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GLib	#pylint: disable=wrong-import-position @UnresolvedImport

width = 400
height = 200

def main():
	other = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	other.set_title("Other")
	other.set_size_request(width/2, height/2)
	other.connect("delete_event", Gtk.main_quit)
	other.show_all()

	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_title("Main")
	window.set_size_request(width, height)
	window.connect("delete_event", Gtk.main_quit)
	vbox = Gtk.VBox(homogeneous=False, spacing=0)
	def add_button(title, callback):
		hbox = Gtk.HBox(homogeneous=False, spacing=0)
		vbox.pack_start(hbox, expand=False, fill=False, padding=10)
		btn = Gtk.Button(label=title)
		hbox.pack_start(btn, expand=False, fill=False, padding=10)
		def on_clicked(*_args):
			def after_delay():
				print("**********************************************\nCALLING %s" % callback)
				callback()
				print("DONE\n**********************************************")
			GLib.timeout_add(5*1000, after_delay)
		btn.connect("clicked", on_clicked)
	add_button("raise Main in 5 seconds", window.present)
	add_button("raise Other in 5 seconds", other.present)
	window.add(vbox)
	window.show_all()
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import dbus.service
from typing import Any

from xpra.notifications.common import parse_image_data, parse_image_path
from xpra.dbus.helper import dbus_to_native
from xpra.util.str_fn import csv
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("dbus", "notify")

BUS_NAME = "org.freedesktop.Notifications"
BUS_PATH = "/org/freedesktop/Notifications"

ACTIONS = envbool("XPRA_NOTIFICATIONS_ACTIONS", True)


def parse_hints(dbus_hints) -> dict:
    hints = {}
    h = dbus_to_native(dbus_hints)
    for x in ("image-data", "icon_data"):
        data = h.pop(x, None)
        if data:
            v = parse_image_data(data)
            if v:
                hints["image-data"] = v
                break
    if "image-data" not in hints:
        image_path = h.pop("image-path", "")
        v = parse_image_path(image_path)
        if v:
            hints["image-data"] = v
    for x in ("action-icons", "category", "desktop-entry", "resident", "transient", "x", "y", "urgency"):
        v = h.get(x)
        if v is not None:
            hints[x] = v
    log("parse_hints(%s)=%s", dbus_hints, hints)
    return hints


def get_capabilities() -> list[str]:
    caps = ["body", "icon-static"]
    if ACTIONS:
        caps += ["actions", "action-icons"]
    return caps


class DBUSNotificationsForwarder(dbus.service.Object):
    """
    We register this class as handling notifications on the session dbus,
    optionally replacing an existing instance if one exists.

    The generalized callback signatures are:
     notify_callback(dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout)
     close_callback(nid)
    """

    def __init__(self, bus, notify_callback=None, close_callback=None):
        self.bus = bus
        self.notify_callback = notify_callback
        self.close_callback = close_callback
        self.active_notifications = set()
        self.counter = 0
        self.dbus_id = os.environ.get("DBUS_SESSION_BUS_ADDRESS", "")
        bus_name = dbus.service.BusName(BUS_NAME, bus=bus)
        super().__init__(bus_name, BUS_PATH)

    def get_info(self) -> dict[str, Any]:
        return {
            "active": tuple(self.active_notifications),
            "counter": self.counter,
            "dbus-id": self.dbus_id,
            "bus-name": BUS_NAME,
            "bus-path": BUS_PATH,
            "capabilities": get_capabilities(),
        }

    def next_id(self) -> int:
        self.counter += 1
        return self.counter

    @dbus.service.method(BUS_NAME, in_signature='susssasa{sv}i', out_signature='u')
    def Notify(self, app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout):
        if replaces_nid == 0:
            nid = self.next_id()
        else:
            nid = int(replaces_nid)
        log("Notify%s nid=%s, counter=%i, callback=%s",
            (app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout),
            nid, self.counter, self.notify_callback)
        self.active_notifications.add(nid)
        if self.notify_callback:
            try:
                actions = tuple(str(x) for x in actions)
                hints = parse_hints(hints)
                args = (
                    self.dbus_id, int(nid), str(app_name),
                    int(replaces_nid), str(app_icon),
                    str(summary), str(body),
                    actions, hints, int(expire_timeout),
                )
            except Exception as e:
                log("Notify(..)", exc_info=True)
                log.error("Error: failed to parse Notify arguments:")
                log.estr(e)
                return 0
            with log.trap_error("Error calling notification handler"):
                self.notify_callback(*args)
        log("Notify returning %s", nid)
        return nid

    @dbus.service.method(BUS_NAME, out_signature='ssss')
    def GetServerInformation(self):
        # name, vendor, version, spec-version
        from xpra import __version__
        v = ["xpra-notification-proxy", "xpra", __version__, "0.9"]
        log("GetServerInformation()=%s", v)
        return v

    @dbus.service.method(BUS_NAME, out_signature='as')
    def GetCapabilities(self):
        caps = get_capabilities()
        log("GetCapabilities()=%s", csv(caps))
        return caps

    @dbus.service.method(BUS_NAME, in_signature='u')
    def CloseNotification(self, nid) -> None:
        log("CloseNotification(%s) callback=%s", nid, self.close_callback)
        try:
            self.active_notifications.remove(int(nid))
        except KeyError:
            return
        else:
            if self.close_callback:
                self.close_callback(nid)
            self.NotificationClosed(nid, 3)  # 3="The notification was closed by a call to CloseNotification"

    def is_notification_active(self, nid) -> bool:
        return nid in self.active_notifications

    @dbus.service.signal(BUS_NAME, signature='uu')
    def NotificationClosed(self, nid, reason) -> None:
        log(f"NotificationClosed({nid}, {reason})")

    @dbus.service.signal(BUS_NAME, signature='us')
    def ActionInvoked(self, nid, action_key) -> None:
        log(f"ActionInvoked({nid}, {action_key})")

    def release(self) -> None:
        try:
            self.bus.release_name(BUS_NAME)
        except dbus.exceptions.DBusException as e:
            log("release()", exc_info=True)
            log.error("Error releasing the dbus notification forwarder:")
            for x in str(e).split(": "):
                log.estr(x)

    def __str__(self):
        return f"DBUS-NotificationsForwarder({BUS_NAME})"


def register(notify_callback=None, close_callback=None, replace=False):
    from xpra.dbus.common import init_session_bus
    bus = init_session_bus()
    flags = dbus.bus.NAME_FLAG_DO_NOT_QUEUE
    if replace:
        flags |= dbus.bus.NAME_FLAG_REPLACE_EXISTING
    request = bus.request_name(BUS_NAME, flags)
    log(f"notifications: bus name {BUS_NAME!r}, request={request}")
    if request == dbus.bus.REQUEST_NAME_REPLY_EXISTS:
        raise ValueError(f"the name {BUS_NAME!r} is already claimed on the session bus")
    return DBUSNotificationsForwarder(bus, notify_callback, close_callback)


def main():
    register()
    from xpra.os_util import gi_import
    glib = gi_import("GLib")
    mainloop = glib.MainLoop()
    mainloop.run()


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stdint cimport uint8_t, uint32_t
from xpra.buffers.membuf cimport MemBuf, getbuf
from typing import Tuple

from xpra.common import SizedBuffer
from xpra.log import Logger
log = Logger("brotli")


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "brotli/encode.h":
    ctypedef enum BrotliEncoderMode:
        BROTLI_MODE_GENERIC
        BROTLI_MODE_TEXT
        BROTLI_MODE_FONT
    ctypedef enum BrotliEncoderOperation:
        BROTLI_OPERATION_PROCESS
        BROTLI_OPERATION_FLUSH
        BROTLI_OPERATION_FINISH
        BROTLI_OPERATION_EMIT_METADATA
    ctypedef enum BrotliEncoderParameter:
        BROTLI_PARAM_MODE
        BROTLI_PARAM_QUALITY
        BROTLI_PARAM_LGWIN
        BROTLI_PARAM_LGBLOCK
        BROTLI_PARAM_DISABLE_LITERAL_CONTEXT_MODELING
        BROTLI_PARAM_SIZE_HINT
        BROTLI_PARAM_LARGE_WINDOW
        BROTLI_PARAM_NPOSTFIX
        BROTLI_PARAM_NDIRECT
        BROTLI_PARAM_STREAM_OFFSET

    ctypedef void BrotliEncoderState
    ctypedef void* brotli_alloc_func
    ctypedef void* brotli_free_func
    ctypedef int BROTLI_BOOL

    uint32_t BrotliEncoderVersion()

    BrotliEncoderState* BrotliEncoderCreateInstance(brotli_alloc_func alloc_func,
                                                    brotli_free_func free_func,
                                                    void* opaque)
    void BrotliEncoderDestroyInstance(BrotliEncoderState* state)

    size_t BrotliEncoderMaxCompressedSize(size_t input_size)

    BROTLI_BOOL BrotliEncoderCompress(int quality, int lgwin,
                                      BrotliEncoderMode mode, size_t input_size,
                                      const uint8_t *input_buffer,
                                      size_t* encoded_size,
                                      uint8_t *encoded_buffer) nogil

    BROTLI_BOOL BrotliEncoderCompressStream(BrotliEncoderState* state,
                                            BrotliEncoderOperation op, size_t* available_in,
                                            const uint8_t** next_in, size_t* available_out, uint8_t** next_out,
                                            size_t* total_out) nogil

    BROTLI_BOOL BrotliEncoderIsFinished(BrotliEncoderState* state)
    BROTLI_BOOL BrotliEncoderHasMoreOutput(BrotliEncoderState* state)


def get_version() -> Tuple[int, int, int]:
    cdef uint32_t bv = BrotliEncoderVersion()
    cdef unsigned int major = bv >> 24
    cdef unsigned int minor = (bv >> 12) & 0xFFF
    cdef unsigned int patch = bv & 0xFFF
    return (major, minor, patch)


DEF BROTLI_MIN_QUALITY = 0
DEF BROTLI_MAX_QUALITY = 11
DEF BROTLI_DEFAULT_WINDOW = 22


def compress(data, int quality=1) -> memoryview:
    #clamp to >2 so that we can use BrotliEncoderMaxCompressedSize:
    quality = max(2, min(BROTLI_MAX_QUALITY, quality))
    cdef size_t max_size = BrotliEncoderMaxCompressedSize(len(data))

    cdef MemBuf out_buf = getbuf(max_size, True)
    cdef uint8_t *out = <uint8_t *> out_buf.get_mem()

    cdef Py_buffer in_buf
    if PyObject_GetBuffer(data, &in_buf, PyBUF_ANY_CONTIGUOUS):
        raise ValueError(f"failed to read data from {type(data)}")
    cdef const uint8_t *in_ptr = <const uint8_t*> in_buf.buf

    cdef size_t out_size = max_size
    cdef int r
    log("brotli.compress(%i bytes, %i) into %i byte buffer", in_buf.len, quality, out_size)
    try:
        with nogil:
            r = BrotliEncoderCompress(quality, BROTLI_DEFAULT_WINDOW,
                                      BROTLI_MODE_GENERIC,
                                      in_buf.len, in_ptr,
                                      &out_size, out)
    finally:
        PyBuffer_Release(&in_buf)
    if not r:
        raise ValueError(f"brotli compression failed: {r}")
    return memoryview(out[:out_size]).toreadonly()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import math
import cairo

DEFAULT_COLOURS = ((0.8, 0, 0), (0, 0, 0.8), (0.1, 0.65, 0.1), (0, 0.6, 0.6), (0.1, 0.1, 0.1))


def round_up_unit(i: int, rounding=10) -> int:
    v = 1
    while v * rounding < i:
        v = v * rounding
    for x in range(10):
        if v * x > i:
            return v * x
    return v * rounding


def make_graph_imagesurface(data, labels=None, width=320, height=200, title=None,
                            show_y_scale=True, show_x_scale=False,
                            min_y_scale=None, rounding=10,
                            start_x_offset=0.0,
                            colours=DEFAULT_COLOURS, dots=False, curves=True) -> cairo.ImageSurface:
    # print("make_graph_pixmap(%s, %s, %s, %s, %s, %s, %s, %s, %s)" % (data, labels, width, height, title,
    #                  show_y_scale, show_x_scale, min_y_scale, colours))
    surface = cairo.ImageSurface(cairo.Format.RGB24, width, height)  # pylint: disable=no-member
    y_label_chars = 4
    x_offset = y_label_chars * 8
    y_offset = 20
    over = 2
    radius = 2
    # inner dimensions (used for graph only)
    w = width - x_offset
    h = height - y_offset * 2
    context = cairo.Context(surface)  # pylint: disable=no-member
    # fill with white:
    context.rectangle(0, 0, width, height)
    context.set_source_rgb(1, 1, 1)
    context.fill()
    # find ranges:
    max_y = 0
    max_x = 0
    for line_data in data:
        x = 0
        for y in line_data:
            if y is not None:
                max_y = max(max_y, y)
            x += 1
            max_x = max(max_x, x)
    # round up the scales:
    scale_x = max_x
    if min_y_scale is not None:
        max_y = max(max_y, min_y_scale)
    scale_y = round_up_unit(max_y, rounding)
    # use black:
    context.set_source_rgb(0, 0, 0)
    # border:
    context.move_to(0, 0)
    context.line_to(width, 0)
    context.line_to(width, height)
    context.line_to(0, height)
    context.line_to(0, 0)
    context.stroke()
    # show vertical line:
    context.move_to(x_offset, y_offset - over)
    context.line_to(x_offset, height - y_offset + over)
    # show horizontal line:
    context.move_to(x_offset - over, height - y_offset)
    context.line_to(width, height - y_offset)
    # units:
    context.select_font_face('Sans')
    context.set_font_size(10)
    # scales
    for i in range(0, 11):
        if show_y_scale:
            context.set_source_rgb(0, 0, 0)
            context.set_line_width(1)
            # vertical:
            y = height - y_offset - h * i / 10
            # text
            if scale_y < 10:
                unit = str(int(scale_y * i) / 10.0)
            else:
                unit = str(int(scale_y * i / 10))
            context.move_to(x_offset - 3 - (x_offset - 6) / y_label_chars * min(y_label_chars, len(unit)), y + 3)
            context.show_text(unit)
            # line indicator
            context.move_to(x_offset - over, y)
            context.line_to(x_offset + over, y)
            context.stroke()
            context.move_to(x_offset + over, y)
            context.set_source_rgb(0.5, 0.5, 0.5)
            context.set_line_width(0.5)
            context.set_dash([3.0, 3.0])
            context.line_to(width, y)
            context.stroke()
            context.set_dash([])
        if show_x_scale:
            context.set_source_rgb(0, 0, 0)
            context.set_line_width(1)
            # horizontal:
            x = x_offset + w * i / 10
            # text
            context.move_to(x - 2, height - 2)
            unit = str(int(scale_x * i / 10))
            context.show_text(unit)
            # line indicator
            context.move_to(x, height - y_offset - over)
            context.line_to(x, height - y_offset + over)
            context.stroke()
    # title:
    if title:
        context.set_source_rgb(0.2, 0.2, 0.2)
        context.select_font_face('Serif')
        context.set_font_size(14)
        context.move_to(x_offset + w / 2 - len(title) * 14 / 2, 14)
        context.show_text(title)
        context.stroke()
    # now draw the actual data, clipped to the graph region:
    context.save()
    context.new_path()
    context.set_line_width(0.0)
    context.rectangle(x_offset, y_offset, x_offset + w, y_offset + h)
    context.clip()
    context.set_line_width(1.5)
    for i, line_data in enumerate(data):
        colour = colours[i % len(colours)]
        context.set_source_rgb(*colour)
        j = 0
        last_v = (-1, -1, -1)
        for v in line_data:
            x = x_offset + w * (j - start_x_offset) / (max(1, max_x - 2))
            if v is not None:
                if max_y > 0:
                    y = height - y_offset - h * v / scale_y
                else:
                    y = 0
                if last_v != (-1, -1, -1):
                    lx, ly = last_v[1:3]
                    if curves:
                        x1 = (lx * 2 + x) / 3
                        y1 = ly
                        x2 = (lx + x * 2) / 3
                        y2 = y
                        context.curve_to(x1, y1, x2, y2, x, y)
                        context.stroke()
                    else:
                        context.line_to(x, y)
                        context.stroke()
                if dots:
                    context.arc(x, y, radius, 0, 2 * math.pi)
                    context.fill()
                    context.stroke()
                    context.move_to(x, y)
                else:
                    context.move_to(x, y)
                last_v = v, x, y
            j += 1
        context.stroke()
    context.restore()
    for i, line_data in enumerate(data):
        # show label:
        if labels and len(labels) > i:
            label = labels[i]
            colour = colours[i % len(colours)]
            context.set_source_rgb(*colour)
            context.select_font_face('Serif')
            context.set_font_size(12)
            context.move_to(x_offset / 2 + (width - x_offset) * i / len(labels), height - 4)
            context.show_text(label)
            context.stroke()
    return surface



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any

from xpra.common import noerr
from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.util.system import SIGNAMES
from xpra.util.str_fn import bytestostr
from xpra.exit_codes import ExitCode, ExitValue
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.signals import install_signal_handlers
from xpra.gtk.css_overrides import inject_css_overrides
from xpra.client.base.client import XpraClientBase
from xpra.client.base.command import InfoXpraClient
from xpra.platform.gui import force_focus
from xpra.gtk.dialogs.qrcode import qr_pixbuf
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("client", "util")

IPV6 = envbool("XPRA_IPV6", False)

inject_css_overrides()


def dpath(caps: typedict, *path) -> Any:
    d: typedict = caps
    for x in path:
        val = d.dictget(x)
        if not val:
            return None
        d = typedict(val)
    return d


class QRCodeClient(InfoXpraClient):

    def do_command(self, caps: typedict) -> None:
        sockets = dpath(caps, "network", "sockets")
        if not sockets:
            log.error("Error: network.sockets path not found in server info response")
            super().quit(ExitCode.PACKET_FAILURE)
            return
        # this will also prevent timeouts:
        self._protocol.close()
        self.exit_code = 0
        addr_types: dict[tuple[str, int], list[str]] = {}
        for socktype in ("ws", "wss"):
            sockdefs = typedict(sockets.dictget(socktype, {}))
            addresses = sockdefs.tupleget("addresses", ())
            for address in addresses:
                try:
                    host, port = address
                except (ValueError, IndexError):
                    continue
                host = bytestostr(host)
                if host.startswith("127.0.0.") or host.startswith("::1"):
                    continue
                if host.find(":") >= 0:
                    if IPV6:
                        host = "[%s]" % host
                    else:
                        continue
                addr_types.setdefault((host, port), []).append(socktype)
        log("addr_types=%s", addr_types)
        if addr_types:
            uris = []
            for addr, socktypes in addr_types.items():
                host, port = addr
                proto = "http" if "ws" in socktypes else "https"
                if (proto == "http" and port == 80) or (proto == "https" and port == 443):
                    uri = "%s://%s/" % (proto, host)
                else:
                    uri = "%s://%s:%i/" % (proto, host, port)
                uris.append(uri)

            def show_addresses():
                w = QRCodeWindow(uris)
                w.show_all()

            GLib.idle_add(show_addresses)
        else:
            noerr(sys.stdout.write, "no addresses found")
            noerr(sys.stdout.flush)
            super().quit(ExitCode.OK)

    def exit_loop(self) -> None:
        Gtk.main_quit()

    def quit(self, exit_code: ExitValue = 0) -> None:
        # only exit if we encountered an error
        # InfoXpraClient calls quit(ExitCode.OK) on `connection-lost`,
        # but we don't want to exit then
        if exit_code != 0:
            super().quit(exit_code)

    def run(self) -> ExitValue:
        # override so we can use a GTK main loop instead
        XpraClientBase.run(self)
        Gtk.main()
        return self.exit_code


class QRCodeWindow(Gtk.Window):

    def __init__(self, uris):
        self.exit_code = None
        super().__init__(type=Gtk.WindowType.TOPLEVEL)
        self.connect("delete_event", self.exit)
        title = "Xpra Server QR Codes"
        n = len(uris)
        self.set_title(title)
        self.set_size_request(512 * n, 580)
        self.set_position(Gtk.WindowPosition.CENTER)
        icon = get_icon_pixbuf("xpra.png")
        if icon:
            self.set_icon(icon)
        hbox = Gtk.HBox(homogeneous=True, spacing=20)
        for uri in uris:
            vbox = Gtk.VBox(homogeneous=False, spacing=10)
            vbox.add(label(uri))
            pixbuf = qr_pixbuf(uri, width=360, height=360)
            image = Gtk.Image().new_from_pixbuf(pixbuf)
            vbox.add(image)
            hbox.add(vbox)
        self.add(hbox)
        add_close_accel(self, self.exit)
        install_signal_handlers("", self.handle_signal)

    def handle_signal(self, signum, frame=None):
        log("handle_signal(%s, %s)", SIGNAMES.get(signum, signum), frame)
        self.exit_code = 128 - (signum or 0)
        GLib.idle_add(self.exit)

    def run(self) -> ExitValue:
        self.show_all()
        force_focus()
        self.present()
        if Gtk.main_level() == 0:
            Gtk.main()
        return self.exit_code or 0

    def exit(self, *args):
        log("exit%s calling %s", args, Gtk.main_quit)
        if self.exit_code is None:
            self.exit_code = 0
        Gtk.main_quit()


def do_main(opts):
    import os
    if os.environ.get("XPRA_HIDE_DOCK") is None:
        os.environ["XPRA_HIDE_DOCK"] = "1"
    from xpra.platform import program_context
    with program_context("qrcode", "QRCode"):
        Gtk.Window.set_auto_startup_notification(setting=False)
        c = QRCodeClient(opts)
        # add_close_accel(w, Gtk.main_quit)
        return c.run()


def main(_args):
    from xpra.scripts.config import make_defaults_struct
    defaults = make_defaults_struct()
    return do_main(defaults)


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import binascii
import re
from enum import Enum
from typing import Iterable
from collections.abc import Callable, Mapping

from xpra.util.env import envbool


def std(v, extras="-+,./: ") -> str:
    def f(c):
        return str.isalnum(c) or c in extras

    return "".join(filter(f, bytestostr(v or "")))


def alnum(v: str | bytes) -> str:
    s = bytestostr(v)
    # noinspection PyTypeChecker
    return "".join(filter(str.isalnum, s))


def nonl(x) -> str:
    if not x:
        return ""
    return str(x).replace("\n", "\\n").replace("\r", "\\r")


def obsc(v) -> str:
    OBSCURE_PASSWORDS = envbool("XPRA_OBSCURE_PASSWORDS", True)
    if OBSCURE_PASSWORDS:
        return "".join("*" for _ in (bytestostr(v) or ""))
    return v


def convert(text: str) -> float | str:
    return float(text) if text.isdigit() else text


def alphanum(key: str) -> list:
    return [convert(c) for c in re.split(r'([-+]?\d+\.?\d*)', key)]


def sort_human(lst: list[str]) -> list:
    lst.sort(key=alphanum)
    return lst


def sorted_nicely(items: Iterable) -> Iterable[str]:
    """ Sort the given iterable in the way that humans expect."""

    def alphanum_key(key) -> list:
        return [convert(c) for c in re.split(r"(\d+)", bytestostr(key))]

    return sorted(items, key=alphanum_key)


def csv(v: Iterable) -> str:
    try:
        return ", ".join(str(x) for x in v)
    except TypeError:
        return str(v)


def is_valid_hostname(hostname: str) -> bool:
    if len(hostname) > 255:
        return False
    if hostname[-1] == ".":
        hostname = hostname[:-1]  # strip exactly one dot from the right, if present
    allowed = re.compile(r"(?!-)[A-Z\d-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(x) for x in hostname.split("."))


class Ellipsizer:
    __slots__ = ("obj", "limit")

    def __init__(self, obj, limit=100):
        self.obj = obj
        self.limit = limit

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if self.obj is None:
            return "None"
        return repr_ellipsized(self.obj, self.limit)


def repr_ellipsized(obj, limit=100) -> str:
    if isinstance(obj, str):
        s = repr(obj)
        if len(s) > limit > 6:
            return s[:limit // 2 - 2] + " .. " + s[2 - limit // 2:]
        return s
    if isinstance(obj, memoryview):
        obj = obj.tobytes()
    if isinstance(obj, bytes):
        try:
            s = repr(obj)
        except TypeError:
            s = binascii.hexlify(obj).decode()
        if len(s) > limit > 6:
            return s[:limit // 2 - 2] + " .. " + s[2 - limit // 2:]
        return s
    return repr_ellipsized(repr(obj), limit)


def print_nested_dict(d: Mapping, prefix: str = "", lchar: str = "*", pad: int = 32,
                      vformat=None, print_fn: Callable | None = None,
                      version_keys=("version", "revision"), hex_keys=("data",)) -> None:
    # "smart" value formatting function:
    def sprint(arg) -> None:
        if print_fn:
            print_fn(arg)
        else:
            print(arg)

    def vf(k, v) -> str:
        if vformat:
            fmt = vformat
            if isinstance(vformat, dict):
                fmt = vformat.get(k)
            if fmt is not None:
                return nonl(fmt(v))
        if isinstance(k, str):
            try:
                if any(k.find(x) >= 0 for x in version_keys):
                    return nonl(pver(v)).lstrip("v")
                if any(k.find(x) >= 0 for x in hex_keys):
                    return binascii.hexlify(v).decode("latin1")
            except TypeError:
                pass
        return nonl(pver(v, ", ", ", "))

    indent = pad - len(prefix) - len(lchar)
    for k in sorted_nicely(d.keys()):
        v = d[k]
        if isinstance(v, dict):
            nokey = v.get("", (v.get(None)))
            if nokey is not None:
                sprint("%s%s %s : %s" % (prefix, lchar, bytestostr(k).ljust(indent), vf(k, nokey)))
                for x in ("", None):
                    v.pop(x, None)
            else:
                sprint("%s%s %s" % (prefix, lchar, bytestostr(k)))
            print_nested_dict(v, prefix + "  ", "-", vformat=vformat, print_fn=print_fn,
                              version_keys=version_keys, hex_keys=hex_keys)
        else:
            sprint("%s%s %s : %s" % (prefix, lchar, bytestostr(k).ljust(indent), vf(k, v)))


def nicestr(obj) -> str:
    """ Python 3.10 and older don't give us a nice string representation for enums """
    if isinstance(obj, Enum):
        return str(obj.value)
    return str(obj)


def strtobytes(x) -> bytes:
    if isinstance(x, bytes):
        return x
    try:
        return str(x).encode("latin1")
    except UnicodeEncodeError:
        return str(x).encode("utf8")


def bytestostr(x) -> str:
    if isinstance(x, (bytes, bytearray)):
        return x.decode("latin1")
    return str(x)


def hexstr(v) -> str:
    return bytestostr(binascii.hexlify(memoryview_to_bytes(v)))


def decode_str(x, try_encoding="utf8") -> str:
    """
    When we want to decode something (usually a byte string) no matter what.
    Try with utf8 first then fallback to just bytestostr().
    """
    try:
        return x.decode(try_encoding)
    except (AttributeError, UnicodeDecodeError):
        return bytestostr(x)


def pver(v, numsep: str = ".", strsep: str = ", ") -> str:
    # print for lists with version numbers, or CSV strings
    if isinstance(v, (list, tuple)):
        types = list(set(type(x) for x in v))
        if len(types) == 1:
            if types[0] is int:
                return numsep.join(str(x) for x in v)
            if types[0] is str:
                return strsep.join(str(x) for x in v)
            if types[0] is bytes:
                def s(x) -> str:
                    try:
                        return x.decode("utf8")
                    except UnicodeDecodeError:
                        return bytestostr(x)

                return strsep.join(s(x) for x in v)
    return bytestostr(v)


def memoryview_to_bytes(v) -> bytes:
    if isinstance(v, bytes):
        return v
    if isinstance(v, memoryview):
        return v.tobytes()
    if isinstance(v, bytearray):
        return bytes(v)
    return strtobytes(v)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from typing import Any
from collections.abc import Iterable

# default implementation uses pycups
from xpra.common import noop
from xpra.platform import platform_import
from xpra.util.str_fn import print_nested_dict, pver
from xpra.util.env import envbool
from xpra.log import Logger, consume_verbose_argv

log = Logger("printing")

RAW_MODE = envbool("XPRA_PRINTER_RAW", False)


def get_printers() -> dict[str, Any]:
    return {}


def get_printer_attributes(_name: str) -> list:
    return []


def get_default_printer() -> str:
    return ""


def print_files(printer: str, filenames: Iterable[str], title: str, options: dict):
    raise RuntimeError("no print implementation available")


def printing_finished(_printpid) -> bool:
    return True


def init_printing(printers_modified_callback=noop) -> None:  # pylint: disable=unused-argument
    """ overridden in platform code """


def cleanup_printing() -> None:
    """ overridden in platform code """


DEFAULT_MIMETYPES = ["application/pdf", "application/postscript"]

MIMETYPES: list[str] | None = None


def get_mimetypes() -> list[str]:
    global MIMETYPES
    if MIMETYPES is None:
        v = os.environ.get("XPRA_PRINTING_MIMETYPES", )
        if v is not None:
            MIMETYPES = v.split(",")
        else:
            MIMETYPES = DEFAULT_MIMETYPES
        if RAW_MODE:
            MIMETYPES.append("raw")
        # make it easier to test different mimetypes:
        PREFERRED_MIMETYPE = os.environ.get("XPRA_PRINTING_PREFERRED_MIMETYPE")
        if PREFERRED_MIMETYPE:
            if PREFERRED_MIMETYPE in MIMETYPES:
                MIMETYPES.remove(PREFERRED_MIMETYPE)
                MIMETYPES.insert(0, PREFERRED_MIMETYPE)
            else:
                log.warn("Warning: ignoring invalid preferred printing mimetype: %s", PREFERRED_MIMETYPE)
                log.warn(" allowed mimetypes: %s", MIMETYPES)
    log("get_mimetype()=%s", MIMETYPES)
    return MIMETYPES


def get_info() -> dict[str, Any]:
    return default_get_info()


def default_get_info() -> dict[str, Any]:
    return {
        "mimetypes": {
            "": get_mimetypes(),
            "default": DEFAULT_MIMETYPES,
        }
    }


platform_import(globals(), "printing", False,
                "init_printing",
                "cleanup_printing",
                "get_printers",
                "get_default_printer",
                "print_files",
                "printing_finished",
                "get_info",
                "DEFAULT_MIMETYPES")


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    consume_verbose_argv(argv, "printing")
    from xpra.util.str_fn import nonl

    def dump_dict(d):
        pk = None
        try:
            for pk, pv in d.items():
                try:
                    if isinstance(pv, bytes):
                        sv = pv.decode("utf8")
                    else:
                        sv = nonl(pver(pv))
                except Exception:
                    sv = repr(pv)
                print(f"        {pk:32} : {sv}")
        except Exception as e:
            print(f"        error on {pk}: {e}")
            print(f"        raw attributes: {d}")

    def dump_info(d):
        print("System Configuration:")
        print_nested_dict(d)

    def dump_printers(d):
        for k in sorted(d.keys()):
            v = d[k]
            print("Printers:")
            print(f"* {k}")
            dump_dict(v)
            attr = get_printer_attributes(k)
            if attr:
                print(" attributes:")
                for a in attr:
                    print(f"        {a}")

    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.util.str_fn import csv
    with program_context("Printing", "Printing"):
        enable_color()
        try:
            init_printing()
        except Exception as e:
            print("Error: initializing the printing tool")
            print(f" {e}")
        if len(argv) <= 1:
            dump_printers(get_printers())
            print("")
            dump_info(get_info())
            return 0
        printers = get_printers()
        if not printers:
            print("Cannot print: no printers found")
            return 1
        if len(argv) == 2:
            filename = argv[1]
            if not os.path.exists(filename):
                print(f"Cannot print file {filename!r}: file does not exist")
                return 1
            printer = get_default_printer()  # pylint: disable=assignment-from-none
            if not printer:
                printer = list(printers.keys())[0]
                if len(printers) > 1:
                    print("More than one printer found: " + csv(printer.keys()))
            print(f"Using printer {printer!r}")
            filenames = [filename]
        if len(argv) > 2:
            printer = argv[1]
            if printer not in printers:
                print(f"Invalid printer {printer!r}")
                return 1
            filenames = argv[2:]
            for filename in filenames:
                if not os.path.exists(filename):
                    print(f"File {filename!r} does not exist")
                    return 1
        print("Printing: " + csv(filenames))
        print_files(printer, filenames, "Print Command", {})
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
from collections.abc import Callable, Sequence

from xpra.common import noop
from xpra.util.str_fn import repr_ellipsized
from xpra.util.env import envbool, IgnoreWarningsContext
from xpra.os_util import OSX, gi_import
from xpra.codecs.icon_util import INKSCAPE_RE
from xpra.gtk.widget import scaled_image, menuitem
from xpra.gtk.pixbuf import get_pixbuf_from_data
from xpra.gtk.dialogs.about import about, close_about
from xpra.platform.gui import get_icon_size
from xpra.platform.paths import get_icon_dir
from xpra.log import Logger

log = Logger("menu")

Gtk = gi_import("Gtk")
GdkPixbuf = gi_import("GdkPixbuf")

MENU_ICONS = envbool("XPRA_MENU_ICONS", True)
HIDE_DISABLED_MENU_ENTRIES = envbool("XPRA_HIDE_DISABLED_MENU_ENTRIES", False)

LOSSLESS = "Lossless"
QUALITY_OPTIONS_COMMON: dict[int, str] = {
    50: "Average",
    30: "Low",
}
MIN_QUALITY_OPTIONS = QUALITY_OPTIONS_COMMON.copy()
MIN_QUALITY_OPTIONS |= {
    0: "None",
    75: "High",
}
MIN_QUALITY_OPTIONS = dict(sorted(MIN_QUALITY_OPTIONS.items()))
QUALITY_OPTIONS = QUALITY_OPTIONS_COMMON.copy()
QUALITY_OPTIONS |= {
    0: "Auto",
    1: "Lowest",
    90: "Best",
    100: LOSSLESS,
}
QUALITY_OPTIONS = dict(sorted(QUALITY_OPTIONS.items()))

SPEED_OPTIONS_COMMON: dict[int, str] = {
    70: "Low Latency",
    50: "Average",
    30: "Low Bandwidth",
}
MIN_SPEED_OPTIONS = SPEED_OPTIONS_COMMON.copy()
MIN_SPEED_OPTIONS[0] = "None"
MIN_SPEED_OPTIONS = dict(sorted(MIN_SPEED_OPTIONS.items()))
SPEED_OPTIONS = SPEED_OPTIONS_COMMON.copy()
SPEED_OPTIONS |= {
    0: "Auto",
    1: "Lowest Bandwidth",
    100: "Lowest Latency",
}
SPEED_OPTIONS = dict(sorted(SPEED_OPTIONS.items()))


def get_bandwidth_menu_options() -> Sequence[int]:
    options = []
    for x in os.environ.get("XPRA_BANDWIDTH_MENU_OPTIONS", "1,2,5,10,20,50,100").split(","):
        try:
            options.append(int(float(x) * 1000 * 1000))
        except ValueError:
            log.warn("Warning: invalid bandwidth menu option '%s'", x)
    return options


BANDWIDTH_MENU_OPTIONS = get_bandwidth_menu_options()


def ll(m) -> str:
    try:
        return "%s:%s" % (type(m), m.get_label())
    except AttributeError:
        return str(m)


def set_sensitive(widget, sensitive: bool) -> None:
    if OSX:
        if sensitive:
            widget.show()
        else:
            widget.hide()
    widget.set_sensitive(sensitive)


def load_pixbuf(data) -> GdkPixbuf.Pixbuf:
    loader = GdkPixbuf.PixbufLoader()
    loader.write(data)
    loader.close()
    return loader.get_pixbuf()


def get_appimage(app_name: str, icondata=b"", menu_icon_size=24) -> Gtk.Image | None:
    pixbuf = None
    if app_name and not icondata:
        # try to load from our icons:
        nstr = app_name.lower()
        icon_filename = os.path.join(get_icon_dir(), "%s.png" % nstr)
        if os.path.exists(icon_filename):
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename=icon_filename)

    def err(e) -> None:
        log("failed to load icon", exc_info=True)
        log.error(f"Error: failed to load icon data for {app_name!r}:")
        log.estr(e)
        log.error(" data=%s", repr_ellipsized(icondata))

    if not pixbuf and icondata:
        # gtk pixbuf loader:
        try:
            pixbuf = load_pixbuf(icondata)
        except Exception as e:
            log("pixbuf loader failed", exc_info=True)
            if re.findall(INKSCAPE_RE, icondata):
                try:
                    pixbuf = load_pixbuf(re.sub(INKSCAPE_RE, b"", icondata))
                except Exception:
                    # there is almost no chance pillow will be able to load it
                    # (it doesn't even have svg support at time of writing)
                    # so don't bother showing another error for the same data:
                    err(e)
                    icondata = None
    if not pixbuf and icondata:
        # let's try pillow:
        try:
            from xpra.codecs.pillow.decoder import open_only, is_svg  # pylint: disable=import-outside-toplevel
            if not is_svg(icondata):
                img = open_only(icondata)
                has_alpha = img.mode == "RGBA"
                width, height = img.size
                rowstride = width * (3 + int(has_alpha))
                pixbuf = get_pixbuf_from_data(img.tobytes(), has_alpha, width, height, rowstride)
                return scaled_image(pixbuf, icon_size=menu_icon_size)
        except Exception as e:
            err(e)
    if pixbuf:
        return scaled_image(pixbuf, icon_size=menu_icon_size)
    return None


# utility method to ensure there is always only one CheckMenuItem
# selected in a submenu:
def ensure_item_selected(submenu, item, recurse=True):
    if not isinstance(item, Gtk.CheckMenuItem):
        return None
    if item.get_active():
        # deactivate all except this one
        def deactivate(items, skip=None) -> None:
            for x in items:
                if x == skip:
                    continue
                if isinstance(x, Gtk.MenuItem):
                    submenu = x.get_submenu()
                    if submenu and recurse:
                        deactivate(submenu.get_children(), skip)
                if isinstance(x, Gtk.CheckMenuItem):
                    if x != item and x.get_active():
                        x.set_active(False)

        deactivate(submenu.get_children(), item)
        return item

    # ensure there is at least one other active item

    def get_active_item(items) -> int:
        for x in items:
            if isinstance(x, Gtk.MenuItem):
                submenu = x.get_submenu()
                if submenu:
                    a = get_active_item(submenu.get_children())
                    if a:
                        return a
            if isinstance(x, Gtk.CheckMenuItem) and x.get_active():
                return x
        return 0

    active = get_active_item(submenu.get_children())
    if active:
        return active
    # if not then keep this one active:
    item.set_active(True)
    return item


def make_min_auto_menu(title, min_options, options,
                       get_current_min_value,
                       get_current_value,
                       set_min_value_cb,
                       set_value_cb):
    # note: we must keep references to the parameters on the submenu
    # (closures and gtk callbacks don't mix so well!)
    submenu = Gtk.Menu()
    submenu.get_current_min_value = get_current_min_value
    submenu.get_current_value = get_current_value
    submenu.set_min_value_cb = set_min_value_cb
    submenu.set_value_cb = set_value_cb
    fstitle = Gtk.MenuItem()
    fstitle.set_label(f"Fixed {title}")
    set_sensitive(fstitle, False)
    submenu.append(fstitle)
    submenu.menu_items = {}
    submenu.min_menu_items = {}
    submenu.updating = False

    def populate_menu(options, value, set_fn):
        found_match = False
        items = {}
        if value and value > 0 and value not in options:
            options[value] = f"{value}%"
        for s in sorted(options.keys()):
            t = options.get(s)
            qi = Gtk.CheckMenuItem(label=t)
            qi.set_draw_as_radio(True)
            candidate_match = s >= max(0, value)
            qi.set_active(not found_match and candidate_match)
            found_match |= candidate_match
            qi.connect('activate', set_fn, submenu)
            if s > 0:
                qi.set_tooltip_text(f"{s}%")
            submenu.append(qi)
            items[s] = qi
        return items

    def set_value(item, ss) -> None:
        if not item.get_active() or submenu.updating:
            return
        try:
            submenu.updating = True
            # user selected a new value from the menu:
            s = -1
            for ts, tl in options.items():
                if tl == item.get_label():
                    s = ts
                    break
            if s >= 0 and s != ss.get_current_value():
                log("setting %s to %s", title, s)
                ss.set_value_cb(s)
                # deselect other items:
                for x in ss.menu_items.values():
                    if x != item:
                        x.set_active(False)
                # min is only relevant in auto-mode:
                if s != 0:
                    for v, x in ss.min_menu_items.items():
                        x.set_active(v == 0)
        finally:
            submenu.updating = False

    submenu.menu_items.update(populate_menu(options, get_current_value(), set_value))
    submenu.append(Gtk.SeparatorMenuItem())
    mstitle = Gtk.MenuItem()
    mstitle.set_label(f"Minimum {title}:")
    set_sensitive(mstitle, False)
    submenu.append(mstitle)

    def set_min_value(item, ss) -> None:
        if not item.get_active() or submenu.updating:
            return
        try:
            submenu.updating = True
            # user selected a new min-value from the menu:
            s = -1
            for ts, tl in min_options.items():
                if tl == item.get_label():
                    s = ts
                    break
            if s >= 0 and s != ss.get_current_min_value():
                log(f"setting min-{title} to {s}")
                ss.set_min_value_cb(s)
                # deselect other min items:
                for x in ss.min_menu_items.values():
                    if x != item:
                        x.set_active(False)
                # min requires auto-mode:
                for x in ss.menu_items.values():
                    if x.get_label() == "Auto":
                        if not x.get_active():
                            x.activate()
                    else:
                        x.set_active(False)
        finally:
            submenu.updating = False

    mv = -1
    if get_current_value() <= 0:
        mv = get_current_min_value()
    submenu.min_menu_items.update(populate_menu(min_options, mv, set_min_value))
    submenu.show_all()
    return submenu


def make_encodingsmenu(get_current_encoding: Callable, set_encoding: Callable,
                       encodings, server_encodings) -> Gtk.Menu:
    encodings_submenu = Gtk.Menu()
    populate_encodingsmenu(encodings_submenu, get_current_encoding, set_encoding, encodings, server_encodings)
    return encodings_submenu


def populate_encodingsmenu(encodings_submenu, get_current_encoding: Callable, set_encoding: Callable,
                           encodings, server_encodings) -> None:
    from xpra.codecs.loader import get_encoding_help, get_encoding_name
    encodings_submenu.get_current_encoding = get_current_encoding
    encodings_submenu.set_encoding = set_encoding
    encodings_submenu.encodings = encodings
    encodings_submenu.server_encodings = server_encodings
    encodings_submenu.index_to_encoding = {}
    encodings_submenu.encoding_to_index = {}
    NAME_TO_ENCODING = {}
    for i, encoding in enumerate(encodings):
        if encoding == "-":
            encodings_submenu.append(Gtk.SeparatorMenuItem())
            continue
        if encoding.startswith("label:"):
            if not HIDE_DISABLED_MENU_ENTRIES:
                label = Gtk.MenuItem()
                label.set_label(encoding[len("label:"):])
                set_sensitive(label, False)
                encodings_submenu.append(label)
            continue
        name = get_encoding_name(encoding)
        descr = get_encoding_help(encoding)
        NAME_TO_ENCODING[name] = encoding
        encoding_item = Gtk.CheckMenuItem(label=name)
        if descr:
            if encoding not in server_encodings:
                descr += "\n(not available on this server)"
            encoding_item.set_tooltip_text(descr)

        def encoding_changed(item):
            ensure_item_selected(encodings_submenu, item)
            enc = NAME_TO_ENCODING.get(item.get_label())
            log("encoding_changed(%s) enc=%s, current=%s", item, enc, encodings_submenu.get_current_encoding())
            if enc is not None and encodings_submenu.get_current_encoding() != enc:
                encodings_submenu.set_encoding(enc)

        log("populate_encodingsmenu(..) encoding=%s, current=%s, active=%s",
            encoding, get_current_encoding(), encoding == get_current_encoding())
        encoding_item.set_active(encoding == get_current_encoding())
        sensitive = encoding in server_encodings
        if not sensitive and HIDE_DISABLED_MENU_ENTRIES:
            continue
        set_sensitive(encoding_item, encoding in server_encodings)
        encoding_item.set_draw_as_radio(True)
        encoding_item.connect("toggled", encoding_changed)
        encodings_submenu.append(encoding_item)
        encodings_submenu.index_to_encoding[i] = encoding
        encodings_submenu.encoding_to_index[encoding] = i
    encodings_submenu.show_all()


# pylint: disable=import-outside-toplevel
class MenuHelper:

    def __init__(self, client):
        self.menu = None
        self.menu_shown = False
        self.menu_icon_size = get_icon_size()
        self.handshake_menuitem: Callable = self.do_handshake_menuitem
        self.set_client(client)

    def set_client(self, client) -> None:
        if client:
            self.client = client

            def shortcut() -> None:
                self.handshake_menuitem = self.menuitem

            client.after_handshake(shortcut)

    def build(self):
        log(f"build() menu={self.menu}")
        if self.menu is None:
            try:
                self.menu = self.setup_menu()
            except Exception as e:
                log("build()", exc_info=True)
                log.error("Error: failed to setup menu")
                log.estr(e)
        return self.menu

    def setup_menu(self):
        log("setup_menu()")
        return self.do_setup_menu(self.get_menu_items())

    def do_setup_menu(self, items: Sequence[Gtk.ImageMenuItem | Gtk.MenuItem]):
        menu = Gtk.Menu()
        for menu_item in items:
            menu.append(menu_item)
        menu.connect("deactivate", self.menu_deactivated)
        menu.show_all()
        return menu

    def get_menu_items(self) -> list[Gtk.ImageMenuItem | Gtk.MenuItem]:
        raise NotImplementedError()

    def cleanup(self) -> None:
        self.close_menu()
        close_about()

    def close_menu(self) -> None:
        if self.menu_shown:
            self.menu.popdown()
            self.menu_shown = False

    def menu_deactivated(self, *args) -> None:
        log(f"menu_deactivated{args}")
        self.menu_shown = False

    def activate(self, button=1, time=0) -> None:
        log("activate(%s, %s)", button, time)
        self.show_menu(button, time)

    def popup(self, button: int, time) -> None:
        log("popup(%s, %s)", button, time)
        self.show_menu(button, time)

    def show_menu(self, button: int, time) -> None:
        self.close_menu()
        if not self.menu:
            log.warn("Warning: menu is not available yet")
            return
        with IgnoreWarningsContext():
            self.menu.popup(None, None, None, None, button, time)
        self.menu_shown = True

    def show_shortcuts(self, *args) -> None:
        self.client.show_shorcuts(*args)

    def show_session_info(self, *args) -> None:
        self.client.show_session_info(*args)

    def show_bug_report(self, *args) -> None:
        self.client.show_bug_report(*args)

    def get_image(self, icon_name, size=None):
        return self.client.get_image(icon_name, size)

    def after_handshake(self, cb: Callable, *args) -> None:
        if self.client:
            self.client.after_handshake(cb, *args)

    def do_handshake_menuitem(self, *args, **kwargs) -> Gtk.ImageMenuItem:
        """ Same as menuitem() but this one will be disabled until we complete the server handshake """
        mi = self.menuitem(*args, **kwargs)
        set_sensitive(mi, False)

        def enable_menuitem(*_args) -> None:
            set_sensitive(mi, True)

        self.after_handshake(enable_menuitem)
        return mi

    def make_menu(self) -> Gtk.Menu:
        return Gtk.Menu()

    def menuitem(self, title, icon_name="", tooltip="", cb: Callable = noop, **kwargs) -> Gtk.ImageMenuItem:
        """ Utility method for easily creating an ImageMenuItem """
        image = None
        if MENU_ICONS:
            image = kwargs.get("image")
            if icon_name and not image:
                icon_size = self.menu_icon_size or get_icon_size()
                image = self.get_image(icon_name, icon_size)

        # Gtk adds ImageMenuItem as arguments to the callback, but we don't want any:
        def menu_cb(*_args) -> None:
            if cb:
                cb()

        return menuitem(title, image, tooltip, menu_cb)

    def checkitem(self, title, cb: Callable = noop, active=False) -> Gtk.CheckMenuItem:
        """ Utility method for easily creating a CheckMenuItem """
        check_item = Gtk.CheckMenuItem(label=title)
        check_item.set_active(active)
        if cb:
            check_item.connect("toggled", cb)
        check_item.show()
        return check_item

    def make_aboutmenuitem(self) -> Gtk.ImageMenuItem:
        return self.menuitem("About Xpra", "xpra.png", cb=about)

    def make_updatecheckmenuitem(self) -> Gtk.ImageMenuItem:
        def show_update_window() -> None:
            from xpra.gtk.dialogs.update_status import get_update_status_window
            w = get_update_status_window()
            w.show()
            w.check()

        return self.menuitem("Check for updates", "update.png", cb=show_update_window)

    def make_qrmenuitem(self) -> Gtk.ImageMenuItem:
        try:
            from xpra.net.qrcode.qrencode import encode_image
        except ImportError as e:
            log(f"no qrcode support {e}")
            return None
        from xpra.gtk.dialogs.qrcode import show_qr

        def show() -> None:
            uri = self.client.display_desc.get("display_name")
            show_qr(uri)

        qr_menuitem = self.menuitem("Show QR connection string", "qr.png", cb=show)
        log("make_qrmenuitem() qrencode.encode_image=%s", encode_image)
        if encode_image:
            def with_connection(*_args) -> None:
                uri = self.client.display_desc.get("display_name")
                if not uri or not any(uri.startswith(proto) for proto in ("tcp:", "ws:", "wss:")):
                    set_sensitive(qr_menuitem, False)
                    qr_menuitem.set_tooltip_text("server uri is not shareable")

            self.after_handshake(with_connection)
        else:
            set_sensitive(qr_menuitem, False)
            qr_menuitem.set_tooltip_text("qrencode library is missing")
        return qr_menuitem

    def make_sessioninfomenuitem(self) -> Gtk.ImageMenuItem:
        def show_session_info_cb(*_args) -> None:
            # we define a generic callback to remove the arguments
            # (which contain the menu widget and are of no interest to the 'show_session_info' function)
            self.show_session_info()

        sessioninfomenuitem = self.handshake_menuitem("Session Info", "statistics.png", cb=show_session_info_cb)
        return sessioninfomenuitem

    def make_bugreportmenuitem(self) -> Gtk.ImageMenuItem:
        return self.menuitem("Bug Report", "bugs.png", cb=self.show_bug_report)

    def make_docsmenuitem(self) -> Gtk.ImageMenuItem:
        from xpra.scripts.main import show_docs
        from xpra.scripts.config import find_docs_path
        docs_menuitem = self.menuitem("Documentation", "documentation.png", cb=show_docs)
        if not find_docs_path():
            docs_menuitem.set_tooltip_text("documentation not found!")
            set_sensitive(docs_menuitem, False)
        return docs_menuitem

    def make_html5menuitem(self) -> Gtk.ImageMenuItem:
        def show_html5() -> None:
            from xpra.scripts.main import run_html5
            from xpra.util.thread import start_thread
            url_options = {}
            try:
                for k in ("port", "host", "username", "mode", "display"):
                    v = self.client.display_desc.get(k)
                    if v is not None:
                        url_options[k] = v
            except Exception:
                pass
            start_thread(run_html5, "open HTML5 client", True, args=(url_options,))

        from xpra.scripts.config import find_html5_path
        html5_menuitem = self.menuitem("HTML5 client", "browser.png", cb=show_html5)
        if not find_html5_path():
            html5_menuitem.set_tooltip_text("html5 client not found!")
            set_sensitive(html5_menuitem, False)
        return html5_menuitem

    def make_closemenuitem(self) -> Gtk.ImageMenuItem:
        return self.menuitem("Close Menu", "close.png", cb=self.close_menu)



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import signal
import threading

from libc.stdint cimport uintptr_t
from xpra.x11.bindings.xlib cimport (
    Display, XOpenDisplay,
    XNextEvent, XEvent, XErrorEvent,
    XSetIOErrorHandler, XSetErrorHandler,
)


cdef int exit_code = -1
exit_event = threading.Event()


def err(s) -> None:
    try:
        sys.stderr.write("%s\n"  % s)
        sys.stderr.flush()
    except IOError:
        pass


cdef void end(msg, int code = exit_code) noexcept:
    global exit_code, exit_event
    err(msg)
    exit_code = code
    exit_event.set()


cdef int x11_io_error_handler(Display *display) except 0:
    global message
    message = b"X11 fatal IO error"
    exit_code = 0
    return 0


cdef int x11_error_handler(Display *display, XErrorEvent *event) except 0:
    #X11 error handler called (ignored)
    return 0


SIGNAMES = {}
for signame in (sig for sig in dir(signal) if sig.startswith("SIG") and not sig.startswith("SIG_")):
    SIGNAMES[getattr(signal, signame)] = signame


def os_signal(signum, _frame=None) -> None:
    end("\ngot signal %s" % SIGNAMES.get(signum, signum), 128-signum)


def xwait(display_name: str="") -> None:
    global exit_code
    if not display_name:
        end("no display specified", 1)
        return
    bname = display_name.encode()
    cdef char* name = bname
    cdef Display * d = XOpenDisplay(name)
    if d==NULL:
        end("failed to open display %r" % (display_name, ), 1)
        return
    #we have successfully connected to the display,
    #so from now on return success on exit:
    XSetErrorHandler(&x11_error_handler)
    XSetIOErrorHandler(&x11_io_error_handler)
    cdef XEvent e
    while True:
        with nogil:
            XNextEvent(d, &e)


def main(args):
    signal.signal(signal.SIGINT, os_signal)
    signal.signal(signal.SIGTERM, os_signal)

    display_name = os.environ.get("DISPLAY") if not args else args[0]
    args = (display_name, )
    t = threading.Thread(target=xwait, name="xwait", args=args, daemon=True)
    t.start()

    exit_event.wait()
    sys.exit(exit_code)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from ctypes import CDLL, c_char_p, c_int

# This definition of the error codes was found here:
# https://social.technet.microsoft.com/wiki/contents/articles/37870.remote-desktop-client-troubleshooting-disconnect-codes-and-reasons.aspx
ERROR_CODES: dict[int, str] = {
    -1: "RDP client thread ended",
    0: "No error",
    1: "User-initiated client disconnect.",
    2: "User-initiated client logoff.",
    3: "Your Remote Desktop Services session has ended, possibly for one of the following reasons:  The administrator has ended the session. An error occurred while the connection was being established. A network problem occurred.  For help solving the problem, see \"Remote Desktop\" in Help and Support.",
    260: "Remote Desktop can't find the computer . This might mean that does not belong to the specified network.  Verify the computer name and domain that you are trying to connect to.",
    262: "This computer can't connect to the remote computer.  Your computer does not have enough virtual memory available. Close your other programs, and then try connecting again. If the problem continues, contact your network administrator or technical support.",
    264: "This computer can't connect to the remote computer.  The two computers couldn't connect in the amount of time allotted. Try connecting again. If the problem continues, contact your network administrator or technical support.",
    266: "The smart card service is not running. Please start the smart card service and try again.",
    516: "Remote Desktop can't connect to the remote computer for one of these reasons:  1) Remote access to the server is not enabled 2) The remote computer is turned off 3) The remote computer is not available on the network  Make sure the remote computer is turned on and connected to the network, and that remote access is enabled.",
    522: "A smart card reader was not detected. Please attach a smart card reader and try again.",
    772: "This computer can't connect to the remote computer.  The connection was lost due to a network error. Try connecting again. If the problem continues, contact your network administrator or technical support.",
    778: "There is no card inserted in the smart card reader. Please insert your smart card and try again.",
    1030: "Because of a security error, the client could not connect to the remote computer. Verify that you are logged on to the network, and then try connecting again.",
    1032: "The specified computer name contains invalid characters. Please verify the name and try again.",
    1034: "An error has occurred in the smart card subsystem. Please contact your helpdesk about this error.",
    1796: "This computer can't connect to the remote computer.  Try connecting again. If the problem continues, contact the owner of the remote computer or your network administrator.",
    1800: "Your computer could not connect to another console session on the remote computer because you already have a console session in progress.",
    2056: "The remote computer disconnected the session because of an error in the licensing protocol. Please try connecting to the remote computer again or contact your server administrator.",
    2308: "Your Remote Desktop Services session has ended.  The connection to the remote computer was lost, possibly due to network connectivity problems. Try connecting to the remote computer again. If the problem continues, contact your network administrator or technical support.",
    2311: "The connection has been terminated because an unexpected server authentication certificate was received from the remote computer. Try connecting again. If the problem continues, contact the owner of the remote computer or your network administrator.",
    2312: "A licensing error occurred while the client was attempting to connect (Licensing timed out). Please try connecting to the remote computer again.",
    2567: "The specified username does not exist. Verify the username and try logging in again. If the problem continues, contact your system administrator or technical support.",
    2820: "This computer can't connect to the remote computer.  An error occurred that prevented the connection. Try connecting again. If the problem continues, contact the owner of the remote computer or your network administrator.",
    2822: "Because of an error in data encryption, this session will end. Please try connecting to the remote computer again.",
    2823: "The user account is currently disabled and cannot be used. For assistance, contact your system administrator or technical support.",
    2825: "The remote computer requires Network Level Authentication, which your computer does not support. For assistance, contact your system administrator or technical support.",
    3079: "A user account restriction (for example, a time-of-day restriction) is preventing you from logging on. For assistance, contact your system administrator or technical support.",
    3080: "The remote session was disconnected because of a decompression failure at the client side. Please try connecting to the remote computer again.",
    3335: "As a security precaution, the user account has been locked because there were too many logon attempts or password change attempts. Wait a while before trying again, or contact your system administrator or technical support.",
    3337: "The security policy of your computer requires you to type a password on the Windows Security dialog box. However, the remote computer you want to connect to cannot recognize credentials supplied using the Windows Security dialog box. For assistance, contact your system administrator or technical support.",
    3590: "The client can't connect because it doesn't support FIPS encryption level.  Please lower the server side required security level Policy, or contact your network administrator for assistance",
    3591: "This user account has expired. For assistance, contact your system administrator or technical support.",
    3592: "Failed to reconnect to your remote session. Please try to connect again.",
    3593: "The remote PC doesn't support Restricted Administration mode.",
    3847: "This user account's password has expired. The password must change in order to logon. Please update the password or contact your system administrator or technical support.",
    3848: "A connection will not be made because credentials may not be sent to the remote computer. For assistance, contact your system administrator.",
    4103: "The system administrator has restricted the times during which you may log in. Try logging in later. If the problem continues, contact your system administrator or technical support.",
    4104: "The remote session was disconnected because your computer is running low on video resources.  Close your other programs, and then try connecting again. If the problem continues, contact your network administrator or technical support.",
    4359: "The system administrator has limited the computers you can log on with. Try logging on at a different computer. If the problem continues, contact your system administrator or technical support.",
    4615: "You must change your password before logging on the first time. Please update your password or contact your system administrator or technical support.",
    4871: "The system administrator has restricted the types of logon (network or interactive) that you may use. For assistance, contact your system administrator or technical support.",
    5127: "The Kerberos sub-protocol User2User is required. For assistance, contact your system administrator or technical support.",
    6919: "Remote Desktop cannot connect to the remote computer because the authentication certificate received from the remote computer is expired or invalid.  In some cases, this error might also be caused by a large time discrepancy between the client and server computers.",
    7431: "Remote Desktop cannot verify the identity of the remote computer because there is a time or date difference between your computer and the remote computer. Make sure your computer's clock is set to the correct time, and then try connecting again. If the problem occurs again, contact your network administrator or the owner of the remote computer.",
    8711: "Your computer can't connect to the remote computer because your smart card is locked out. Contact your network administrator about unlocking your smart card or resetting your PIN.",
    9479: "Could not auto-reconnect to your applications,please re-launch your applications",
    9732: "Client and server versions do not match. Please upgrade your client software and then try connecting again.",
    33554433: "Failed to reconnect to the remote program. Please restart the remote program.",
    33554434: "The remote computer does not support RemoteApp. For assistance, contact your system administrator.",
    50331649: "Your computer can't connect to the remote computer because the username or password is not valid. Type a valid user name and password.",
    50331650: "Your computer can't connect to the remote computer because it can't verify the certificate revocation list. Contact your network administrator for assistance.",
    50331651: "Your computer can't connect to the remote computer due to one of the following reasons:  1) The requested Remote Desktop Gateway server address and the server SSL certificate subject name do not match. 2) The certificate is expired or revoked. 3) The certificate root authority does not trust the certificate.  Contact your network administrator for assistance.",
    50331652: "Your computer can't connect to the remote computer because the SSL certificate was revoked by the certification authority. Contact your network administrator for assistance.",
    50331653: "This computer can't verify the identity of the RD Gateway. It's not safe to connect to servers that can't be identified. Contact your network administrator for assistance.",
    50331654: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server address requested and the certificate subject name do not match. Contact your network administrator for assistance.",
    50331655: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server's certificate has expired or has been revoked. Contact your network administrator for assistance.",
    50331656: "Your computer can't connect to the remote computer because an error occurred on the remote computer that you want to connect to. Contact your network administrator for assistance.",
    50331657: "An error occurred while sending data to the Remote Desktop Gateway server. The server is temporarily unavailable or a network connection is down. Try again later, or contact your network administrator for assistance.",
    50331658: "An error occurred while receiving data from the Remote Desktop Gateway server. Either the server is temporarily unavailable or a network connection is down. Try again later, or contact your network administrator for assistance.",
    50331659: "Your computer can't connect to the remote computer because an alternate logon method is required. Contact your network administrator for assistance.",
    50331660: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server address is unreachable or incorrect. Type a valid Remote Desktop Gateway server address.",
    50331661: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server is temporarily unavailable. Try reconnecting later or contact your network administrator for assistance.",
    50331662: "Your computer can't connect to the remote computer because the Remote Desktop Services client component is missing or is an incorrect version. Verify that setup was completed successfully, and then try reconnecting later.",
    50331663: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server is running low on server resources and is temporarily unavailable. Try reconnecting later or contact your network administrator for assistance.",
    50331664: "Your computer can't connect to the remote computer because an incorrect version of rpcrt4.dll has been detected. Verify that all components for Remote Desktop Gateway client were installed correctly.",
    50331665: "Your computer can't connect to the remote computer because no smart card service is installed. Install a smart card service and then try again, or contact your network administrator for assistance.",
    50331666: "Your computer can't stay connected to the remote computer because the smart card has been removed. Try again using a valid smart card, or contact your network administrator for assistance.",
    50331667: "Your computer can't connect to the remote computer because no smart card is available. Try again using a smart card.",
    50331668: "Your computer can't stay connected to the remote computer because the smart card has been removed. Reinsert the smart card and then try again.",
    50331669: "Your computer can't connect to the remote computer because the user name or password is not valid. Please type a valid user name and password.",
    50331671: "Your computer can't connect to the remote computer because a security package error occurred in the transport layer. Retry the connection or contact your network administrator for assistance.",
    50331672: "The Remote Desktop Gateway server has ended the connection. Try reconnecting later or contact your network administrator for assistance.",
    50331673: "The Remote Desktop Gateway server administrator has ended the connection. Try reconnecting later or contact your network administrator for assistance.",
    50331674: "Your computer can't connect to the remote computer due to one of the following reasons:   1) Your credentials (the combination of user name, domain, and password) were incorrect. 2) Your smart card was not recognized.",
    50331675: "Remote Desktop can't connect to the remote computer for one of these reasons:  1) Your user account is not listed in the RD Gateway's permission list 2) You might have specified the remote computer in NetBIOS format (for example, computer1), but the RD Gateway is expecting an FQDN or IP address format (for example, computer1.fabrikam.com or 157.60.0.1).  Contact your network administrator for assistance.",
    50331676: "Remote Desktop can't connect to the remote computer for one of these reasons:  1) Your user account is not authorized to access the RD Gateway 2) Your computer is not authorized to access the RD Gateway 3) You are using an incompatible authentication method (for example, the RD Gateway might be expecting a smart card but you provided a password)  Contact your network administrator for assistance.",
    50331679: "Your computer can't connect to the remote computer because your network administrator has restricted access to this RD Gateway server. Contact your network administrator for assistance.",
    50331680: "Your computer can't connect to the remote computer because the web proxy server requires authentication. To allow unauthenticated traffic to an RD Gateway server through your web proxy server, contact your network administrator.",
    50331681: "Your computer can't connect to the remote computer because your password has expired or you must change the password. Please change the password or contact your network administrator or technical support for assistance.",
    50331682: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server reached its maximum allowed connections. Try reconnecting later or contact your network administrator for assistance.",
    50331683: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server does not support the request. Contact your network administrator for assistance.",
    50331684: "Your computer can't connect to the remote computer because the client does not support one of the Remote Desktop Gateway's capabilities. Contact your network administrator for assistance.",
    50331685: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server and this computer are incompatible. Contact your network administrator for assistance.",
    50331686: "Your computer can't connect to the remote computer because the credentials used are not valid. Insert a valid smart card and type a PIN or password, and then try connecting again.",
    50331687: "Your computer can't connect to the remote computer because your computer or device did not pass the Network Access Protection requirements set by your network administrator. Contact your network administrator for assistance.",
    50331688: "Your computer can't connect to the remote computer because no certificate was configured to use at the Remote Desktop Gateway server. Contact your network administrator for assistance.",
    50331689: "Your computer can't connect to the remote computer because the RD Gateway server that you are trying to connect to is not allowed by your computer administrator. If you are the administrator, add this Remote Desktop Gateway server name to the trusted Remote Desktop Gateway server list on your computer and then try connecting again.",
    50331690: "Your computer can't connect to the remote computer because your computer or device did not meet the Network Access Protection requirements set by your network administrator, for one of the following reasons:  1) The Remote Desktop Gateway server name and the server's public key certificate subject name do not match. 2) The certificate has expired or has been revoked. 3) The certificate root authority does not trust the certificate. 4) The certificate key extension does not support encryption. 5) Your computer cannot verify the certificate revocation list.  Contact your network administrator for assistance.",
    50331691: "Your computer can't connect to the remote computer because a user name and password are required to authenticate to the Remote Desktop Gateway server instead of smart card credentials.",
    50331692: "Your computer can't connect to the remote computer because smart card credentials are required to authenticate to the Remote Desktop Gateway server instead of a user name and password.",
    50331693: "Your computer can't connect to the remote computer because no smart card reader is detected. Connect a smart card reader and then try again, or contact your network administrator for assistance.",
    50331695: "Your computer can't connect to the remote computer because authentication to the firewall failed due to missing firewall credentials. To resolve the issue, go to the firewall website that your network administrator recommends, and then try the connection again, or contact your network administrator for assistance.",
    50331696: "Your computer can't connect to the remote computer because authentication to the firewall failed due to invalid firewall credentials. To resolve the issue, go to the firewall website that your network administrator recommends, and then try the connection again, or contact your network administrator for assistance.",
    50331698: "Your Remote Desktop Services session ended because the remote computer didn't receive any input from you.",
    50331699: "The connection has been disconnected because the session timeout limit was reached.",
    50331700: "Your computer can't connect to the remote computer because an invalid cookie was sent to the Remote Desktop Gateway server. Contact your network administrator for assistance.",
    50331701: "Your computer can't connect to the remote computer because the cookie was rejected by the Remote Desktop Gateway server. Contact your network administrator for assistance.",
    50331703: "Your computer can't connect to the remote computer because the Remote Desktop Gateway server is expecting an authentication method different from the one attempted. Contact your network administrator for assistance.",
    50331704: "The RD Gateway connection ended because periodic user authentication failed. Try reconnecting with a correct user name and password. If the reconnection fails, contact your network administrator for further assistance.",
    50331705: "The RD Gateway connection ended because periodic user authorization failed. Try reconnecting with a correct user name and password. If the reconnection fails, contact your network administrator for further assistance.",
    50331707: "Your computer can't connect to the remote computer because the Remote Desktop Gateway and the remote computer are unable to exchange policies. This could happen due to one of the following reasons:     1. The remote computer is not capable of exchanging policies with the Remote Desktop Gateway.     2. The remote computer's configuration does not permit a new connection.     3. The connection between the Remote Desktop Gateway and the remote computer ended.    Contact your network administrator for assistance.",
    50331708: "Your computer can't connect to the remote computer, possibly because the smart card is not valid, the smart card certificate was not found in the certificate store, or the Certificate Propagation service is not running. Contact your network administrator for assistance.",
    50331709: "To use this program or computer, first log on.",
    50331710: "To use this program or computer, you must first log on to an authentication website. Contact your network administrator for assistance.",
    50331711: "Your session has ended. To continue using the program or computer, first log on.",
    50331712: "Your session has ended. To continue using the program or computer, you must first log on to an authentication website. Contact your network administrator for assistance.",
    50331713: "The RD Gateway connection ended because periodic user authorization failed. Your computer or device didn't pass the Network Access Protection (NAP) requirements set by your network administrator. Contact your network administrator for assistance.",
    50331714: "Your computer can't connect to the remote computer because the size of the cookie exceeded the supported size. Contact your network administrator for assistance.",
    50331716: "Your computer can't connect to the remote computer using the specified forward proxy configuration. Contact your network administrator for assistance.",
    50331717: "This computer cannot connect to the remote resource because you do not have permission to this resource. Contact your network administrator for assistance.",
    50331718: "There are currently no resources available to connect to. Retry the connection or contact your network administrator.",
    50331719: "An error occurred while Remote Desktop Connection was accessing this resource. Retry the connection or contact your system administrator.",
    50331721: "Your Remote Desktop Client needs to be updated to the newest version. Contact your system administrator for help installing the update, and then try again.",
    50331722: "Your network configuration doesn't allow the necessary HTTPS ports. Contact your network administrator for help allowing those ports or disabling the web proxy, and then try connecting again.",
    50331723: "We're setting up more resources, and it might take a few minutes. Please try again later.",
    50331724: "The user name you entered does not match the user name used to subscribe to your applications. If you wish to sign in as a different user please choose Sign Out from the Home menu.",
    50331725: "Looks like there are too many users trying out the Azure RemoteApp service at the moment. Please wait a few minutes and then try again.",
    50331726: "Maximum user limit has been reached. Please contact your administrator for further assistance.",
    50331727: "Your trial period for Azure RemoteApp has expired. Ask your admin or tech support for help.",
    50331728: "You no longer have access to Azure RemoteApp. Ask your admin or tech support for help.}",
}


def b(s: str) -> bytes:
    return s.encode("latin1")


def logon(username: str, password: str) -> int:
    # sys.path.insert(0, "E:\\DesktopLogon\\")
    # desktop_logon = ctypes.cdll.LoadLibrary("E:\\DesktopLogon\\DesktopLogon.dll")
    desktop_logon = CDLL("DesktopLogon.dll")
    Logon = desktop_logon.Logon
    Logon.argtypes = [c_char_p, c_char_p, c_char_p, c_char_p]
    Logon.restype = c_int
    s = b("localhost")
    d = b(os.environ.get("USERDOMAIN", "WORKGROUP"))
    u = b(username)
    p = b(password)
    desktop_logon.Logon(s, d, u, p)
    getErrorCode = desktop_logon.getErrorCode
    getErrorCode.argtypes = []
    getErrorCode.restype = c_int
    r = getErrorCode()
    if r:
        from xpra.log import Logger
        log = Logger("win32")
        log.warn(f"Warning: `DesktopLogon` error {r}")
        msg = ERROR_CODES.get(r, "unknown error")
        if msg.find(":") > 0:
            header, msg = msg.split(":", 1)
            log.warn(f" {header}:")
        for line in msg.split("."):
            line = line.strip()
            if line:
                log.warn(f"  {line}")
    return r


def main(argv: list[str]) -> int:
    if len(argv) != 3:
        print("Usage: %s username password")
        return -1
    username = argv[1]
    password = argv[2]
    return logon(username, password)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


import sys


def main():
    from xpra.platform import program_context, command_error
    from xpra.platform.gui import init, set_default_icon
    with program_context("Webcam", "Webcam"):
        from xpra.log import Logger, consume_verbose_argv
        consume_verbose_argv(sys.argv, "webcam")
        log = Logger("webcam")
        set_default_icon("webcam.png")
        init()

        log("importing opencv")
        try:
            import cv2
        except ImportError as e:
            command_error("Error: no opencv support module: %s" % e)
            return 1
        log("cv2=%s", cv2)
        device = 0
        if len(sys.argv) == 2:
            try:
                device = int(sys.argv[1])
            except ValueError:
                command_error("Warning: failed to parse value as a device number: '%s'" % sys.argv[1])
        log("opening %s with device=%s", cv2.VideoCapture, device)  # @UndefinedVariable
        try:
            cap = cv2.VideoCapture(device)  # @UndefinedVariable
        except Exception as e:
            command_error(f"Error: failed to capture video using device {device}:\n{e}")
            return 1
        log.info("capture device for %i: %s", device, cap)
        while True:
            ret, frame = cap.read()
            if not ret:
                command_error("Error: frame capture failed using device %s" % device)
                return 1
            cv2.imshow('frame', frame)  # @UndefinedVariable
            if cv2.waitKey(10) & 0xFF in (ord('q'), 27):  # @UndefinedVariable
                break
        cap.release()
        cv2.destroyAllWindows()
    return 0


if __name__ == "__main__":
    v = main()
    sys.exit(v)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from xpra.client.mixins import logging
from unit.test_util import silence_info
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class MixinsTest(ClientMixinTest):

	def test_logging(self):
		from xpra.log import Logger, is_debug_enabled
		for x in ("network", "crypto"):
			if is_debug_enabled(x):
				#remote logging will be disabled,
				#so we have to skip this test
				return
		opts = AdHocStruct()
		opts.remote_logging = "yes"
		with silence_info(logging):
			self._test_mixin_class(logging.RemoteLogging, opts, {
				"remote-logging" : {"receive"	: True},
				})
		assert len(self.packets)==0
		logger = Logger("util")
		message = "hello"
		logger.info(message)
		assert len(self.packets)==1
		packet = self.packets[0]
		assert packet[0]=="logging", "expected logging packet but got '%s'" % (packet[0],)
		assert packet[1]==20, "expected INFO level (20) but got %s" % (packet[1],)
		#data might be using a compressed wrapper:
		data = getattr(packet[2], "data", packet[2])
		assert data==message, "expected message '%s' but got '%s'" % (message, data)
		#after cleanup, log messages should not be intercepted:
		self.packets = []
		self.mixin.cleanup()
		with silence_info(logging):
			logger.info("foo")
		assert len(self.packets)==0

def main():
	unittest.main()


if __name__ == '__main__':
	main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import Any, Deque
from collections import deque
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.platform.features import CLIPBOARDS
from xpra.net.common import PacketType
from xpra.util.objects import typedict
from xpra.util.env import envint
from xpra.log import Logger

log = Logger("clipboard")

GLib = gi_import("GLib")

MAX_CLIPBOARD_LIMIT = envint("XPRA_CLIPBOARD_LIMIT", 30)
MAX_CLIPBOARD_LIMIT_DURATION = envint("XPRA_CLIPBOARD_LIMIT_DURATION", 3)


class ClipboardConnection(StubSourceMixin):

    PREFIX = "clipboard"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return caps.boolget("clipboard")

    def init_state(self) -> None:
        self.clipboard_enabled = False
        self.clipboard_notifications = False
        self.clipboard_notifications_current = 0
        self.clipboard_notifications_pending = 0
        self.clipboard_progress_timer: int = 0
        self.clipboard_stats: Deque[float] = deque(maxlen=MAX_CLIPBOARD_LIMIT * MAX_CLIPBOARD_LIMIT_DURATION)
        self.clipboard_greedy = False
        self.clipboard_want_targets = False
        self.clipboard_selections = CLIPBOARDS
        self.clipboard_preferred_targets: Sequence[str] = ()

    def cleanup(self) -> None:
        self.cancel_clipboard_progress_timer()

    def parse_client_caps(self, c: typedict) -> None:
        ccaps = c.get(ClipboardConnection.PREFIX)
        if ccaps and isinstance(ccaps, dict):
            ccaps = typedict(ccaps)
            self.clipboard_enabled = ccaps.boolget("enabled", True)
            self.clipboard_notifications = ccaps.boolget("notifications")
            self.clipboard_greedy = ccaps.boolget("greedy")
            self.clipboard_want_targets = ccaps.boolget("want_targets")
            self.clipboard_selections = ccaps.strtupleget("selections", CLIPBOARDS)
            self.clipboard_preferred_targets = ccaps.strtupleget("preferred-targets", ())
        log("client clipboard: enabled=%s, notifications=%s",
            self.clipboard_enabled, self.clipboard_notifications)
        log("client clipboard: greedy=%s, want_targets=%s, selections=%s",
            self.clipboard_greedy, self.clipboard_want_targets, self.clipboard_selections)

    def get_info(self) -> dict[str, Any]:
        return {
            ClipboardConnection.PREFIX: {
                "enabled": self.clipboard_enabled,
                "notifications": self.clipboard_notifications,
                "greedy": self.clipboard_greedy,
                "want-targets": self.clipboard_want_targets,
                "preferred-targets": self.clipboard_preferred_targets,
                "selections": self.clipboard_selections,
            },
        }

    def send_clipboard_enabled(self, reason: str = "") -> None:
        if not self.hello_sent:
            return
        self.send_async("set-clipboard-enabled", self.clipboard_enabled, reason)

    def cancel_clipboard_progress_timer(self) -> None:
        cpt = self.clipboard_progress_timer
        if cpt:
            self.clipboard_progress_timer = 0
            GLib.source_remove(cpt)

    def send_clipboard_progress(self, count: int) -> None:
        if not self.clipboard_notifications or not self.hello_sent or self.clipboard_progress_timer:
            return
        # always set "pending" to the latest value:
        self.clipboard_notifications_pending = count

        # but send the latest value via a timer to tame toggle storms:

        def may_send_progress_update() -> None:
            self.clipboard_progress_timer = 0
            if self.clipboard_notifications_current != self.clipboard_notifications_pending:
                self.clipboard_notifications_current = self.clipboard_notifications_pending
                log("sending clipboard-pending-requests=%s to %s", self.clipboard_notifications_current, self)
                self.send_more("clipboard-pending-requests", self.clipboard_notifications_current)

        delay = (count == 0) * 100
        self.clipboard_progress_timer = GLib.timeout_add(delay, may_send_progress_update)

    def send_clipboard(self, packet) -> None:
        if not self.clipboard_enabled or not self.hello_sent:
            return
        if getattr(self, "suspended", False):
            return
        now = monotonic()
        self.clipboard_stats.append(now)
        if len(self.clipboard_stats) >= MAX_CLIPBOARD_LIMIT:
            event = self.clipboard_stats[-MAX_CLIPBOARD_LIMIT]
            elapsed = now - event
            log("send_clipboard(..) elapsed=%.2f, clipboard_stats=%s", elapsed, self.clipboard_stats)
            if elapsed < 1:
                msg = f"more than {MAX_CLIPBOARD_LIMIT} clipboard requests per second!"
                log.warn("Warning: %s", msg)
                # disable if this rate is sustained for more than S seconds:
                events = [x for x in tuple(self.clipboard_stats) if x > (now - MAX_CLIPBOARD_LIMIT_DURATION)]
                log("%i events in the last %i seconds: %s", len(events), MAX_CLIPBOARD_LIMIT_DURATION, events)
                if len(events) >= MAX_CLIPBOARD_LIMIT * MAX_CLIPBOARD_LIMIT_DURATION:
                    log.warn(" limit sustained for more than %i seconds,", MAX_CLIPBOARD_LIMIT_DURATION)
                return
        # call compress_clibboard via the encode work queue:
        self.queue_encode((True, self.compress_clipboard, (packet,)))

    def compress_clipboard(self, packet: PacketType) -> None:
        # pylint: disable=import-outside-toplevel
        from xpra.net.compression import Compressible, compressed_wrapper
        # Note: this runs in the 'encode' thread!
        packet = list(packet)
        for i, item in enumerate(packet):
            if isinstance(item, Compressible):
                packet[i] = compressed_wrapper(item.datatype, item.data, level=9, can_inline=False, brotli=True)
        self.queue_packet(tuple(packet))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.util.version import get_version_info, get_platform_info, get_host_info
from xpra.util.str_fn import print_nested_dict


def main() -> int:
    from xpra.platform import program_context
    with program_context("Version-Info", "Version Info"):
        print("Build:")
        print_nested_dict(get_version_info())
        print("")
        print("Platform:")
        pi = get_platform_info()
        # ugly workaround for the fact that "sys.platform" has no key..
        if "" in pi:
            pi["sys"] = pi[""]
            del pi[""]
        print_nested_dict(pi)
        print("")
        print("Host:")
        d = get_host_info(2)
        # add os specific version info:
        from xpra.platform.info import get_version_info as pvinfo
        d.update(pvinfo())
        print_nested_dict(d)
    return 0


if __name__ == "__main__":  # pragma: no cover
    main()
